'''
Created on Jan 23, 2021

@author: vladislavkargin
'''
import numpy as np
import matplotlib.pyplot as plt

import ribbons.Ribbon as rb
import ribbons.Tiling as tl

def fibSequence(n, seed = None):
    ''' returns a random increasing sequence of indices (i1, i2, ..., ik)
    such that the indices are between 0 and n-1 and no two indices are 
    adjacent. The number of such sequence is Fibonacci number. The function 
    uses recursion
    '''
    if seed != None:
        np.random.seed(seed)
    seq = []
    if n <= 0:
        return seq
    else:
        x = np.random.randint(2)
        if x == 1: #n-1 is included
            seq = fibSequence(n - 2) + [n - 1]
        else:
            seq = fibSequence(n - 1)
        return seq

def getSquarePairs(sx0, sy0, dir0, sequence):
    ''' generates pairs of squares which should be included in a tiling. 
    parameters: s0x and s0y - coordinates of the original square
    dir0 - direction of original step : 'Down' or 'Right'
    sequence: sequence that regulates, which pairs should be included. '''
    sx = sx0
    sy = sy0
    direction = dir0
    #print(sequence)
    pairs = []
    if sequence == []:
        return pairs
    M = max(sequence) + 1
    for i in range(M):
        if i in sequence:
            if direction == 'Right':
                pairs = pairs + [(sx, sy, 'Right')]
                #print(sx, sy, 'Right') 
            else:
                pairs = pairs + [(sx, sy - 1, 'Up')]
                #print(sx, sy - 1, 'Up')
        if direction == 'Right':
            sx = sx + 1
            sy = sy
            direction = "Down"
        else:
            sx = sx
            sy = sy - 1
            direction = "Right"
    return pairs
    
def pairsForRectangle(M, N, seed = None):
    '''gets all marked pairs for an M-by-N rectangle. It is assumed that N >= M'''
    if seed != None:
        np.random.seed(seed)
    all_pairs =[]
    for i in range(M - 1):
        pairs = getSquarePairs(0, i + 1 , "Down", sequence = fibSequence(2 * i + 2))
        all_pairs = all_pairs + pairs
    for i in range (N - M):
        pairs = getSquarePairs(i, M - 1 , "Right", sequence = fibSequence(2 * M - 1))
        all_pairs = all_pairs + pairs
    for i in range (M - 1):
        pairs = getSquarePairs(N - M + i, M - 1 , "Right", sequence = fibSequence(2 * M - 2 - 2 * i))
        all_pairs = all_pairs + pairs
    return all_pairs

def tilingFromPairs(M, N, pairs):
    ''' creates a tiling of an M-by-N rectangle from a collection of pairs '''
    marked = np.zeros((N,M)) #contains information if a particular cell is in the tiling
    ribbons = []
    pairs_set = set(pairs)
    for y0 in range(M):
        for x0 in range(N):
            if marked[x0, y0] == 0: #this cell is not yet explored.
                marked[x0, y0] = 1
                x = x0
                y = y0
                shape = []
                while (x, y, 'Up') in pairs_set or (x, y, 'Right') in pairs_set:
                    #print("x = ", x, "y = ", y)
                    if (x, y, 'Up') in pairs_set:
                        shape.append(1)
                        y = y + 1
                        marked[x, y] = 1
                    else:
                        shape.append(0)
                        x = x + 1
                        marked[x, y] = 1
                ribbon = rb.Ribbon(x0, y0, shape)
                ribbons.append(ribbon)
    tiling = tl.Tiling(ribbons)
    return tiling
                    
    
    
    
def drawPairs(M, N, pairs):
    '''draws all marked pairs for M-by-N rectangle'''
    fig, ax = plt.subplots()
    ax.set_xlim(0, N + 1)
    ax.set_ylim(0, M + 1)
    for (sx, sy, dr) in pairs:
        if dr == "Up":
            line = plt.Line2D((sx + 0.5, sx + 0.5),(sy + 0.5, sy + 1.5), color = 'black')
            ax.add_line(line)
        else: 
            line = plt.Line2D((sx + 0.5, sx + 1.5),(sy + 0.5, sy + 0.5), color = 'black')
            ax.add_line(line)
    return fig, ax    

def randomStanleyTiling(M, N):
    ''' creates a random ribbon tiling of an M-by-N rectangle using Stanley's 
    algorithm'''
    all_pairs = pairsForRectangle(M, N)
    tiling = tilingFromPairs(M,N, all_pairs)
    return tiling


def lengthCounts(tiling):
    '''creates a dictionary that measures lengths of the ribbons in the tiling. 
    Example: {1:3, 2:5, 7:1} means that there are three 1-ribbons, five two ribbons and 1 7-ribbon.'''
    
    length_counts = {}
    for ribbon in tiling.ribbons:
        l = len(ribbon.shape)
        if l + 1 not in length_counts.keys(): 
            length_counts[l + 1] = 1
        else:
            length_counts[l + 1] = length_counts[l + 1] + 1
    return length_counts    
    

'''
For testing methods
'''
def main():
    n = 20
    seed = 123
    seq = fibSequence(n, seed = seed)
    print(seq)
    
    pairs = getSquarePairs(10, 10, 'Right', seq)
    print(pairs)
    M= 100
    N = 100
    all_pairs = pairsForRectangle(M, N)
    #print("All pairs = ", all_pairs)
    
    drawPairs(M, N, all_pairs)
    
    tiling = tilingFromPairs(M,N, all_pairs)
    #print(tiling)
    tiling.draw(M, N, colormap = 'jet')
    
    length_counts = lengthCounts(tiling)
    print('Length Counts = ', length_counts)

    
    plt.show()

if __name__ == '__main__':
    main()


'''
Created on Jan 17, 2021

@author: vladislavkargin
'''

#import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

class Ribbon(object):
    '''
    classdocs
    '''


    def __init__(self, x, y, shape):
        '''
        creates a ribbon with root square s_{xy} (south-west corner is at (x, y))
        and a list of 0 and 1 that determines the shape. 
        0 means go right, 1 means go up. This list can be empty which will 
        corresponds to a ribbon tile that consist of a single square. 
        '''
        self.x = x 
        self.y = y 
        self.shape = shape
        
    def __str__(self):
        '''
        Returns a string representation of this ribbon.
        return the graph
        '''
        s = '(' + str(self.x) + ',' + str(self.y) + '), ' + str(self.shape)
        return s       
    
    def draw(self, ax = None, block = False, MaxLength = None, colormap = 'prism'):
        if (ax == None):
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        
        
        cmap = cm.get_cmap(colormap)
        if MaxLength == None:
            MaxLength = len(self.shape) + 5
        else: 
            MaxLength = max([MaxLength, len(self.shape)])
        
        c = cmap(len(self.shape)/MaxLength)
        
        rectangle = plt.Rectangle((self.x,self.y), 1, 1, fc=c ,ec="black")
        ax.add_patch(rectangle)
        
        sx = self.x
        sy = self.y
        for i in range(len(self.shape)):
            if self.shape[i] == 0:
                rectangle = plt.Rectangle((sx,sy), 2, 1, fc=c,ec="black")
                ax.add_patch(rectangle)
                if i > 0 and self.shape[i - 1] == 0:
                    line = plt.Line2D((sx, sx),(sy, sy + 1), color = c)
                    ax.add_line(line)
                elif i > 0 and  self.shape[i - 1] == 1:
                    line = plt.Line2D((sx, sx + 1),(sy, sy), color = c)
                    ax.add_line(line) 
                sx = sx + 1 
            else:
                rectangle = plt.Rectangle((sx,sy), 1, 2, fc=c,ec="black")
                ax.add_patch(rectangle) 
                if i > 0 and self.shape[i - 1] == 0:
                    line = plt.Line2D((sx, sx),(sy, sy + 1), color = c)
                    ax.add_line(line)
                elif i > 0 and  self.shape[i - 1] == 1:
                    line = plt.Line2D((sx, sx + 1),(sy, sy), color = c)
                    ax.add_line(line) 
                sy = sy + 1
            
            
        if block:            
            plt.show()
        else:
            plt.draw()
        return fig, ax
'''
For testing methods
'''
def main():
    print('Ribbon methods are used')  
    ribbon = Ribbon(0, 0, [0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1, 1, 1]) 
    _, ax = plt.subplots()
    ax.set_xlim(0, 10)
    ax.set_ylim(0, 10)
    print(ribbon)
    ribbon.draw(ax = ax)
    plt.show()
    
if __name__ == "__main__":
    main() 
    
    


'''
Created on Jan 30, 2021

@author: vladislavkargin
'''
import numpy as np
import matplotlib.pyplot as plt

import ribbons.Ribbon as rb
import ribbons.Tiling as tl
import ribbons.Utility as ut


'''
For testing methods
'''
def main():
    
    M = 10
    N = 100
    tiling = ut.randomStanleyTiling(M, N)
    tiling.draw(M, N, colormap = 'prism')

    
    plt.show()

if __name__ == '__main__':
    main()




'''
Created on Feb 12, 2021

@author: vladislavkargin
'''

''' This is a file for various experiments with Temperley-Lieb algebra '''
import scipy.special as special
import matplotlib.pyplot as plt

import rgs.mndrpy.Pairing as pr
import rgs.mndrpy.ElementTL as tl
import rgs.mndrpy.Utility as ut

def simplify(prng):
    ''' This is the first step to write the pairing in a normal form when 
    we think about the pairing as an element of the Temperley - Lieb algebra
    (see paper of Di Francesco about meander determinant.
    Arguments: prng - a pairing
    Returns: a 3-tuple that consists from index of e_i (which is associated
    with a ringlet, depth of the ringlet 
    (which is needed if we want
    to use this pairing as an encoding of a basis 2 element), and 
    the simplified pairing.
    '''
    N = len(prng.prng)
    #find all ringlets
    ringlets = []
    for i in range(N):
        if prng.prng[i] == i + 1:
            ringlets.append(i)
    #find depths of ringlets (it is easier to do it using the path 
    #representation of the pairing
    path = pr.prng2path(prng)
    x = 0
    y = 0
    S = [(x, y)] #these are the points on the Dyck path.
    for i in range(N):
        if path[i] == 1:
            y += 1
        else:
            x += 1
        S.append((x,y))
    depths = [0] * len(ringlets)
    for i in range(len(ringlets)):
        s = ringlets[i]
        depths[i] = S[s + 1][1] - S[s + 1][0] - 1
    #print("Depths are ", depths)
    #find the first ringlet with the largest depth
    max_depth = max(depths)
    i = depths.index(max_depth)
    x = ringlets[i]
    '''
    depth = 0
    for i in range(len(ringlets)):
        if i == 0 and ringlets[i] > 0:
            x = ringlets[i]
            depth = ringlets[i]
            break
        elif i > 0 and ringlets[i] - ringlets[i - 1] > 2:
            #print(ringlets[i - 1], ringlets[i])
            x = ringlets[i]
            depth = ringlets[i] - ringlets[i-1] - 2
            break
        else: 
            continue
    '''
    #now we calculate the simplified pairing, assuming that 
    #there is a ringlet with positive depth
    if max_depth > 0:
        prng_copy = list(prng.prng)
        y = prng.prng[x - 1]
        prng_copy[x - 1] = x 
        prng_copy[x] = x - 1
        prng_copy[x + 1] = y
        prng_copy[y] = x + 1
        prngS = pr.Pairing(prngArray = prng_copy)
    else:
        prngS = None  
    
    return (ringlets, x, max_depth, prngS)

def normalForm(prng):
    ''' calculate the normal form of the pairing as defined in the 
    paper by DiFrancesco about meander determinants.
    returns: the sequence of the indices of e_i (which is the position of 
    an actionable ringlet) and the depths of the corresponding
    ringlets. The positions are  '''
    nf = []
    prngS = pr.Pairing(prngArray = prng.prng)
    while prngS != None:
        _, x, depth, prngS = simplify(prngS)
        #print(x, depth, prngS)
        if prngS != None:
            nf.append((x,depth))
    return nf

def act(x, prng):
    ''' Acts by e_x on prng. In the picture with Dyck paths this corresponds 
    to adding a square in a valley. A sequence of these actions corresponding
    to the normal form  builds a pairing from the standard pairing.
    Arguments: x - position to act
    prng - prng to act on
    returns: new pairing 
    '''
    #first we need to check that the action is possible 
    N = len(prng.prng)
    if x <= 0 or x >= N - 1:
        print("act says: x should be between 1 and 2n - 2")
        return 
    if prng.prng[x] > x or prng.prng[x + 1] < x + 1:
        print("act says: cannot act by ", x, "on pairing ", str(prng))
        print("x is not a valley in the corresponding Dyck path")
        return 
    prng0 = list(prng.prng)
    y = prng0[x]
    z = prng0[x + 1]
    prng0[y] = z 
    prng0[z] = y
    prng0[x] = x + 1
    prng0[x + 1] = x 
    p = pr.Pairing(prngArray = prng0)
    return p

def act2(x, k, q, el):
    ''' Acts by e_x on an element of TL algebra in the ideal considered
    by Di Francesco.
    In the picture with Dyck paths this corresponds 
    to adding a square in a valley. 
    This is a second variant of the action used by Di Francesco in
     orthogonalization
    Arguments: x - position to act
    k - height at which to act
    q - parameter of the TL algebra
    prng - prng to act on
    returns: an element of a left ideal in TL algebra (see Di Francesco.
    '''
    #TODO
    mu_k = special.eval_chebyu(k - 1, q/2)/special.eval_chebyu(k, q/2)
    #print(mu_k)
    el_new = tl.ElementTL({})
    if isinstance(el, pr.Pairing):
        prng1 = act(x, el)
        el_new = tl.ElementTL({prng1:1}) + tl.ElementTL({el:mu_k})
    elif isinstance(el, tl.ElementTL):
        for prng in el.S.keys():
            el_new = el_new + act2(x,k,q,prng) * el.S[prng]
    else:
        print("act2 says: wrong type of the element")
    return el_new
    


def nf2pairing(nf, n):
    '''takes a normal form and recovers the pairing
    Normal form can be either a sequence of x (locations of valleys in
    the Dyck path representation where we insert a square), or 
    sequence of (x, h_x) where  and h_x is the height of 
    the squares that we insert. In this algorithm we use only 
    the sequence of x-ses.
    arguments: nf - normal form
    n - half-length of the resulting pairing
    returns: the resulting pairing '''
    prng = pr.standardPairing(n)
    if type(nf[-1]) == tuple:
        Xs = list(zip(*nf))[0]
    else:
        Xs = nf
    for x in reversed(Xs):
        prng = act(x, prng)
    return prng


'''
For testing methods
'''
def main():
    pass
    
    widths = [3, 1, 4]
    n = sum(widths)
    prng = pr.rainbows(widths)
    prng.draw()
    print(prng)
    
    path = pr.prng2path(prng)
    ut.plotDyckPath(path)
    '''
    y = 0
    S = [(x, y)]
    for i in range(2 * n):
        if path[i] == 1:
            y += 1
        else:
            x += 1
        S.append((x,y))
    print(S)
    '''    
    ringlets, x, depth, prng1 = simplify(prng)
    print(ringlets, x, depth, prng1)
    if prng1 != None:
        prng1.draw()
    
    
    nf = normalForm(prng)
    print(nf)
    
    
    
    prng = nf2pairing(nf, n)
    
    prng.draw()
    

    
    n = 3 #size of the pairing
    q = 2 #parameter of the algebra
    x = 3 #place to act
    m = 1 
    #vacuum element:
    prng = pr.standardPairing(n) 
    #another element created by acting with e_x on the vacuum element
    prng1 = act(x,prng)
    prng1.draw(asPath = True)
    
    
    '''
    el1 = tl.ElementTL({prng:1}, q = q) #basis element as a part of the TL algebra
    #now we create another element by acting by a different version of e_x on 
    #the vacuum vector. 
    el2 = act2(x, m, q, el1)
    print("Vacuum element: ", el1)
    print(el2)
    #I think the idea is that el2 is orthogonal to el1 and we can also 
    #calculate the norm of el2
    print(el2 | el1) #Problem this is not zero. Have to do it by hands
    print(el2 | el2)
    
    
    #yet another element
    el3 = act2(x - 2, m, q, el2)
    print(el3)

    '''
    plt.show()


if __name__ == '__main__':
    main()


'''
Created on Sep 6, 2020

@author: vladislavkargin
'''

import numpy as np
import matplotlib.pyplot as plt

import pmaps.OrderedTree as ot
import pmaps.TreeUtility as tutil

'''
For testing methods
'''
def main():
    n = 200
    w = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    #w = [1, 0, 1]
    
    '''
    #Calculate the Kolmogorov - Smirnov statistic
    ITER = 10000
    KSsample = []
    for count in range(ITER):
        if (count % 1000 == 0):
            print(count)
        path1 = tutil.randomBridge(n, w, SEED = 0)
        Mplus = max(np.cumsum(path1))
        Mmin = - min(np.cumsum(path1))
        KSsample.append(max([Mplus, Mmin])/np.sqrt(n))
    print(np.mean(KSsample)/np.sqrt(2))
    plt.plot(KSsample)
    '''
    
    '''calculating average height of the trees '''
    ITER = 100
    EHlist0 = [] #predicted expected height
    EHlist = [] #experimental average height
    alpha_range = np.linspace(0.1, 0.95, num = 20, endpoint = False)
    for alpha in alpha_range:
        print("alpha = ", alpha)
        EHlist0.append(np.sqrt((1-alpha)/alpha * np.pi * n))
        k = np.floor(n * alpha)
        H = []
        count = 0
        while count < ITER:
            if count % 100 == 0:
                print("count = ", count)
            try:
                tree = ot.randomSlimTree(k, n, w, SEED = 0)
                h = tree.height()
                H.append(h[0])
                count = count + 1
            except:
                print("there was an error in tree generation at step ", count)
        EH = np.average(H)
        print("EH = ", EH)
        EHlist.append(EH)        
    plt.plot(alpha_range, EHlist, label = "experimental")
    plt.plot(alpha_range, EHlist0, label = "predicted")
    plt.legend()
    plt.grid()
    plt.title("Average tree height, n = " + str(n))
    plt.xlabel("alpha")
    
    #plt.xkcd()
    #plt.xticks()
        #tree.draw()
    
    
    plt.show()
        

if __name__ == '__main__':
    main()


'''
Created on Jan 20, 2021

@author: vladislavkargin
'''

import numpy as np
import matplotlib.pyplot as plt
import ribbons.Ribbon as rb

class Tiling(object):
    '''
    classdocs
    '''


    def __init__(self, ribbons):
        '''
        Constructor
        ribbons is a list of ribbons
        '''
        self.ribbons = ribbons
        
    def __str__(self):
        ''' Returns a string representation of this tiling
        return the graph'''
        s = ''
        for ribbon in self.ribbons:
            s = s + '\n' + str(ribbon)
        return s 
    
    def draw(self, M, N, ax = None, block = False, MaxLength = None, colormap = 'jet'):
        ''' M is the height and N is the width of the plot'''
        if (ax == None):
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
            
        ax.set_xlim(0, N)
        ax.set_ylim(0, M)
        
        if MaxLength == None:
            MN = 3 * int(np.log2(max([M, N])))
        print(MN)
        for ribbon in self.ribbons:
            ribbon.draw(ax, MaxLength = MN, colormap = colormap)
        
        if block:            
            plt.show()
        else:
            plt.draw()
        return fig, ax

''' 
For testing methods
'''
def main():
    print('Ribbon tiling methods are used')  
    ribbon1 = rb.Ribbon(0, 0, [0, 1])
    ribbon2 = rb.Ribbon(0, 1, [1, 0])
    
    tiling = Tiling([ribbon1, ribbon2]) 
    fig, ax = plt.subplots()
    print(tiling)
    tiling.draw(10, 10, ax = ax)
    plt.show()
    
if __name__ == "__main__":
    main() 


'''
Created on Feb 10, 2021

@author: vladislavkargin
'''

import rgs.mndrpy.Pairing as pr
import rgs.mndrpy.Meander as mr

import matplotlib.pyplot as plt


class ElementTL(object):
    '''
    This class represents an element of TL algebra. 
    It can be interpreted in two different ways:
     1) Currently, we think about it as an element of
    the left ideal which generated by the basis elements 
    which have an arbitrary pairing on the left  and a standard pairing
    on the right.
    2) it can also represent an arbitrary element of TL algebra
    if we fold every pairing appropriately.
    
    We realize this object as a dictionary with pairings as keys and 
    its coefficients as values.
    '''


    def __init__(self, S, q = 2):
        '''
        Constructor
        takes a dictionary S that associate a coefficient to a pairing.
        For example, {prng1:1, pring2:3}. This corresponds to element
        prng1 + 2 prng2.
        Another parameter is q. By definition it is set to 2. 
        '''
        self.S = dict.copy(S)
        self.q = q
        
    def __str__(self):
        return ' '.join([str(x) + ":" + str(self.S[x]) for x in self.S.keys()])
    
     
    def __add__(self, other):
        '''adds this element to the other'''
        #make a copy of self
        e = ElementTL(self.S)
        for x in other.S.keys():
            if x in e.S.keys():
                e.S[x] = e.S[x] + other.S[x]
            else:
                e.S[x] = other.S[x]
        return e  
    
    def __sub__(self, other):
        '''subtracts the other element'''
        #make a copy of self
        e = ElementTL(self.S)
        for x in other.S.keys():
            if x in e.S.keys():
                e.S[x] = e.S[x] - other.S[x]
            else:
                e.S[x] = -other.S[x]
        return e
    
    def __mul__(self, other): 
        '''multiplication by a scalar'''
        #make a copy of self
        e = ElementTL(self.S)
        for x in e.S.keys():
            e.S[x] = e.S[x] * other
        return e
    
    def __rmul__(self, other):
        '''multiplication by a scalar. Makes sure that it is 
        OK to multiply on both sides'''
        return self * other

    def __or__(self, other):
        '''This overloading for 'bitwise or' (|) operator is going
        to realize the scalar product of two TL elements.
        arguments: the other TL element
        
        returns the scalar product with the other TL element '''
        #TODO: need to make sure that it uses q-defined scalar product
        if self.q != other.q:
            print("Cannot calculate scalar products of ", str(self), " and ",
                  str(other), ". Parameters q are different in these elements.")
        c = 0
        for x in self.S.keys():
            for y in other.S.keys():
                mndr = mr.Meander(x,y)
                cycles, _ = mndr.findCycles()
                c = c + pow(self.q, len(cycles)) * self.S[x] * other.S[y]
        return c
        


'''
For testing methods
'''
def main():
    
    
    prng1 = pr.Pairing(path = None, prngArray = [3, 2, 1, 0, 5, 4])
    prng2 = pr.Pairing(path = None, prngArray = [1, 0, 3, 2, 5, 4])
    prng3 = pr.Pairing(path = None, prngArray = [5, 2, 1, 4, 3, 0])
    
    mndr = mr.Meander(prng1, prng2)
    print(mndr)
    mndr.draw()
    
    x = ElementTL({prng1:1, prng2:(3)})
    y = ElementTL({prng2:(-1), prng3:0})
    print(x)
    print(y)
    
    print(x + y)
    print(x - y)
    print(x * 3)
    print(3 * x)
    
    print(x|y)
    
    plt.show()
    
if __name__ == '__main__':
    main()    
    


'''
Created on Jul 23, 2020

@author: vladislavkargin
'''

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import cm

import rgs.mndrpy.Pairing as pr
import rgs.mndrpy.Utility as ut


class Meander(object):
    '''
    classdocs
    '''


    def __init__(self, uPairing, dPairing, uprngArray = None,
                    dprngArray = None):
        '''
        Constructor
        Makes the meander from upper and lower pairings
        arguments:
        uPairing, dPairing -- upper and lower parings (as objects or lists)
        
        The other pair of arguments is obsolete (poor design),
        kept only for compatibility with some other functions. 
        uprgnArray, dprngArray -- upper and lower pairings (as lists)
        '''
        if type(uPairing) is list:
            self.uPairing = pr.Pairing(prngArray = uPairing)
        elif uPairing == None:
            self.uPairing = pr.Pairing(prngArray = uprngArray)
        else: 
            self.uPairing = uPairing
            
        if type(dPairing) is list:
            self.dPairing = pr.Pairing(prngArray = dPairing)
        elif uPairing == None:
            self.dPairing = pr.Pairing(prngArray = dprngArray)
        else: 
            self.dPairing = dPairing
        
        
    def __str__(self):
        return str(self.uPairing) + ", " + str(self.dPairing)
    
    def __eq__(self, other):
        '''Overrides the default implementation'''
        if isinstance(other, Meander):
            if self.uPairing == other.uPairing and self.dPairing == other.dPairing: 
                return True
            else: 
                return False
        return False
    
    def __hash__(self):
        return hash(self.uPairing) + 17 * hash(self.dPairing)
    
    def size(self):
        ''' returns n for the meander system on 2n bridges '''
        return self.uPairing.size()
        
    
    def draw(self, ax = None, drawCycles = False, palette = "jet"):
        if ax == None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
        fig = ax.figure
        self.uPairing.draw(ax = ax)
        self.dPairing.draw(ax = ax, up = False)
        if drawCycles == True:
            colormap = cm.get_cmap(palette, 128)
            cycles, _ = self.findCycles()
            for i in range(len(cycles)):
                col = colormap(i/len(cycles))
                #print(i, col)
                self.drawCycle(cycles[i], ax, color = col) 
        return fig, ax
    
    def drawCycle(self, cycle, ax = None, color = "red"):
        if ax == None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            width = len(self.uPairing.prng)
            ax.set_xlim(0, width)
            ax.set_ylim(-width/2, width/2)
        for c, i in enumerate(cycle):
            j = cycle[(c + 1) % len(cycle)]
            xy = (i + (j - i)/2, 0)
            w = np.abs((j - i)/2)
            if c % 2 == 0:
                theta1 = 0.0
                theta2 = 180.0
            else:
                theta1 = 180.0
                theta2 = 360.0
            arc = mpatches.Arc(xy, 2*w, 2*w, 0, theta1, theta2, color = color)
            ax.add_patch(arc)
    
    def isProper(self):
        '''checks if this system has only one cycle'''  
        N = len(self.uPairing.prng)
        #print(N)
        next_v = self.uPairing.prng[0]
        up = False 
        counter = 0    
        while next_v != 0 and counter <= 2*N:
            counter = counter + 1
            if up == True:
                next_v = self.uPairing.prng[next_v]
                up = False
            else:
                next_v = self.dPairing.prng[next_v]
                up = True
        if counter == N - 1:
            return True
        else:
            return False
        
    def  isIrreducible(self):
        '''checks if the meander is irreducible. (It is somewhat different
        then irreducible meanders in the sense of Zvonkin-Lando.)'''
        if self.isProper():
            return True   
        cycles, _ = self.findCycles()
        for cycle in cycles:
            if len(cycle) == max(cycle) - min(cycle) + 1:
                return False
        return True
    
        
    def findCycles(self):
        ''' finds all cycles in the meander and stores them in two structures
        cycles - which is an array of arrays that represent cycles
        cyclesInd - which is an 2n-array with each element marked by 
        the number of the cycle that it belongs to.
        '''
        N = len(self.uPairing.prng)
        c = 0
        up = True #shows which pairing we are going to use in tracking 
        cycles = []
        cyclesInd = [-1] * N
        
        for i in range(N):
            if cyclesInd[i] == -1:
                c = c + 1
                cyclesInd[i] = c
                cycle = [i]
                if up:
                    next_v =  self.uPairing.prng[i]
                    up = False
                else:
                    next_v =  self.dPairing.prng[i]
                    up = True
                #the following cycle can go in infinite loop if one of the
                #pairing is invalid for this reason we break it 
                # if it go for more than N iterations.
                # The result is invalid and likely to break the following programs
                counter = 0
                while (next_v != i and counter <= 2 * N):
                    cyclesInd[next_v] = c
                    cycle.append(next_v)
                    if up:
                        next_v =  self.uPairing.prng[next_v]
                        up = False
                    else:
                        next_v =  self.dPairing.prng[next_v]
                        up = True
                    counter = counter + 1
                cycles.append(cycle)
                if counter > 2 * N:
                    print( "findCycles Warning:")
                    print(" An infinite loop in find components was broken")
                    print(" A likely problem with validity of one of the pairings")
        return cycles, cyclesInd
     
    def calcCycleLengths(self):
        ''' caclulate the lengths of all cycles in the meander'''
        (cycles, _) = self.findCycles()
        lengths = []
        for cycle in cycles:
            lengths.append(len(cycle))
        return lengths
        
         
def randomMeander(n, w = None, seed = None):
    ''' generates a uniformly random meander on half-length n.
    If 2-tuple w != None, generates a meander with pairings distributed 
    according to weights w1 and w2'''
    if seed != None:
        np.random.seed(seed)
    if w == None:
        uP = pr.randomPairing(n)
        dP = pr.randomPairing(n)
    else:
        uP = pr.randomPairing(n, w[0])
        dP = pr.randomPairing(n, w[1])
    mndr = Meander(uP, dP)
    return mndr

def randomCrossSemiMeander(n, seed = None):
    '''generate a random crossing semi-meander'''
    if seed != None:
        np.random.seed(seed)
    uP = pr.randomCrossPairing(n)
    dP = pr.rainbowPairing(n)
    mndr = Meander(uP, dP)
    return mndr
    

def randomComb(n, seed = None):
    ''' generates a random comb meander of half-length n'''
    if seed != None:
        np.random.seed(seed)
    uP = pr.randomPairing(n)
    dP = pr.standardPairing(n)
    mndr = Meander(uP, dP)
    return mndr

def randomSemiMeander(n, seed = None):
    ''' generates a random meander of half-length n that have a random pairing
    on top and the rainbow pairing in the bottom'''
    
    if seed != None:
        np.random.seed(seed)
    uP = pr.randomPairing(n)
    dP = pr.rainbowPairing(n)
    mndr = Meander(uP, dP)
    return mndr

def rainbowMeander(uWidths, dWidths):
    '''
    generates a rainbow meander.
    Parameters:
    uWidths, dWidths -- sizes of the meanders on the top and on the bottom.
    returns a meander
    '''  
    N = sum(uWidths)
    M = sum(dWidths)
    if M != N:
        print("Rainbow meander says: the total sizes of upper and lower meanders")
        print("should be the same. N = ", N, " and M = ", M)
        return 
    uP = pr.rainbows(uWidths)
    dP = pr.rainbows(dWidths)
    return Meander(uP, dP)

def smallRainbowsMeander(amounts, sizes):
    '''
    creates a random meander whose pairings are collections of small rainbows
    '''
    uP = pr.smallRainbows(amounts, sizes)
    dP = pr.smallRainbows(amounts, sizes)
    return Meander(uP, dP)

def drawAsPolygon(mndr):
    ''' draws meander as polygon. The upper pairing is represented
    by a Dyck path above the diagonal y = x, and the lower pairing
    is represented by a Dyck path below the main diagonal. 
    '''
    upath = pr.prng2path(mndr.uPairing)
    ax = ut.plotDyckPath(upath)
    dpath = pr.prng2path(mndr.dPairing)
    ax = ut.plotDyckPath(dpath, ax = ax, method = "lowerLatticePath")
    return ax    
    
'''
For testing methods
'''
def main():
    
    #n = 100
    #seed = None    
    #mndr = randomMeander(n, seed = seed)
    #mndr = randomComb(n, seed = seed)
    #mndr = randomSemiMeander(n, seed = seed)
    #mndr = randomMeander(n, w = ([1,1,1],[1,1,1]), seed = seed)
    #mndr = randomCrossSemiMeander(n)
    amounts = [20, 20]
    sizes = [1, 5]
    mndr = smallRainbowsMeander(amounts, sizes)
    
    print(str(mndr))
    _, ax = mndr.draw(drawCycles = True)
    
    (cycles, cyclesInd) = mndr.findCycles()
    print("Cycles = " + str(cycles))
    print("CyclesInd = " + str(cyclesInd))
    
    lengths = mndr.calcCycleLengths()
    print("Cycle lengths = " + str(lengths))
    
    maxL = max(lengths)
    i = lengths.index(maxL)
    
    print("The length of the largest cycle is " + str(maxL))
    
    mndr.drawCycle(cycles[i], ax = ax)
    
    print("Hash = ", hash(mndr))
    print("Meander? ", mndr.isProper())
    print("Irreducible? ", mndr.isIrreducible())
    
    
    
    ''' Picture for the paper about meanders'''
    
    uWidths = [15, 5, 11]
    dWidths = [18, 13]
    mndr = rainbowMeander(uWidths, dWidths)
    _, ax = mndr.draw(drawCycles = True)  
    (cycles, cyclesInd) = mndr.findCycles()
    print("Cycles = " + str(cycles))
    print("CyclesInd = " + str(cyclesInd))
    lengths = mndr.calcCycleLengths()
    print("Cycle lengths = " + str(lengths))
    maxL = max(lengths)
    i = lengths.index(maxL)
    print("The length of the largest cycle is " + str(maxL))
    mndr.drawCycle(cycles[i], ax = ax)
    
    drawAsPolygon(mndr)
    
    plt.show()


if __name__ == '__main__':
    main()          
        
        


'''
Created on Jan 13, 2020

This is a class for the graph object embedded in a plane. 

The graph is a simple graph - no double edges or loops.
It is a rooted graph, so one of the vertices is chosen as a root. By default it is
vertex 0. 
It is desirable that the root is on the outer boundary, and in all cases when we have outer leafs,
it the root should be an outer leaf. This should helps with resetting the root. 
However, the resetting is not yet implemented.

@author: vladislavkargin
'''

try:
    import matplotlib.pyplot as plt
    import matplotlib.lines as mlines
    import matplotlib.patches as mpatches
    from matplotlib import cm
    possibleToDraw = True
except ImportError:
    possibleToDraw = False
    

import copy
import math
import numpy as np
import numdifftools as nd
import scipy.linalg as la
from scipy.optimize import minimize
from copy import deepcopy

import rgs.pmaps.OrderedTree as ot
import rgs.pmaps.TreeUtility as tu
import rgs.pmaps.Triangulation as tr


class PlanarGraph:
    
    
    #Creates a graph with v vertices and no edges
    def __init__(self, V = 0, root = 0, rootArrow = -1):
        self.V = V
        self.E = 0
        self.root = root
        self.rootArrow = rootArrow #if this is defined then the edge (root, rootArrow) is in
                                #the border of outer country
                                #this is used in the construction of dual graph
        self.adj = []; #list of adjacency lists for vertices
        self.repr = []; #list of coordinates of vertices in a planar representation
        self.labels = []; #list of vertex labels. 
        for _ in range(V):
            self.adj.append([])
            self.repr.append([0., 0.])
        self.dlgr = None #This will be the dual graph. it is calculated on demand via dualGraph method.
                        #(at the least, the graph should be connected at this point).
        self.countries = [] #This is also calculated by dualGraph method.
                            #The country 0 is the outerCountry
        self.R = []  #This is radii of circles in Koebe representation of the graph. It is
        #expensive to calculate them so they are stored.            
        #R -vector of radii for vertices. The first V coordinates are 
        #radii of vertices of the original graph. The next F - 1  coordinates are radii of
        #the vertices of the dual graph
    
    def __str__(self):
        '''
        Returns a string representation of this graph.
        return the number of vertices V, followed by the number
        of edges E, followed by the V adjacency lists
        '''
        s = ("V = " + str(self.V) + ', E = ' + str(self.E) + ", root = " + str(self.root) 
                + ", rootArrow = " + str(self.rootArrow) + '\n')
        for i in range(self.V):
            s = s + str(i) + " -> " + str(self.adj[i]) + '\n'
        return s    
    
    '''
    calculates the degree of vertex v
    '''
    def degree(self, v):
        d = len(self.adj[v])
        return d
    
    def leaves(self):
        '''
        return the list of leaves in the graph
        '''
        lvs = []
        for v in range(self.V):
            if self.degree(v) == 1:
                lvs.append(v)
        return lvs 
    
    def singletons(self):
        '''
        returns the list of singletons (i.e., vertices which are not connected to anything else)
        '''
        sngls = []
        for v in range(self.V):
            if self.degree(v) == 0:
                sngls.append(v)
        return sngls
    
    '''
    returns the edge type. If the edge is to/from leaf, then it is called
    external or outer, and the function returns 0. Otherwise the edge is called
    internal or inner, and the function returns 1.
    NOTE: we do not check if (u, v) is indeed an edge. 
    '''    
    def edgeType(self, u, v):
        if (self.degree(u) == 1 or self.degree(v) == 1):
            return 0
        else:
            return 1
    
    def nextEdge(self, u, v):
        '''
        returns x such that (u, x) is the next edge in the adjacency list of u after (u,v)
        In particular, if u is a leaf, then returns back again v
        '''
        if v not in self.adj[u]:
            print("nextEdge says: the edge (" + str(u) + ", " + str(v) + ") is not in graph")
            return -1
        else: 
            d = self.degree(u)
            i = self.adj[u].index(v)
            x = self.adj[u][(i + 1) % d]
            return x
     
    def previousEdge(self, u, v):
        '''
        returns x such that (u, x) is the previous edge in the adjacency list of u with respect 
        to (u,v)
        In particular, if u is a leaf, then returns back again v
        '''
        if v not in self.adj[u]:
            print("nextEdge says: the edge (" + str(u) + ", " + str(v) + ") is not in graph")
            return -1
        else: 
            d = self.degree(u)
            i = self.adj[u].index(v)
            x = self.adj[u][(i - 1) % d]
            return x 
           
    # Adds the undirected edge u-v to this graph.    
    def addEdge(self, u, v): 
        self.E = self.E + 1
        self.adj[u].append(v)
        self.adj[v].append(u) 
    
    # Adds a directed edge from u to v to this graph.   
    # This graph is meant to be undirected, so this is 
    # a convenience tool to take subgraphs and make sure
    # that the order of adjacency lists is not violated
    def addDirectedEdge(self, u, v):
        if ((v not in self.adj[u]) and (u not in self.adj[v])):
            self.E = self.E + 1
        self.adj[u].append(v)
        

    def addEdgeBeforeAfter(self, u, v, a, b):
        '''
        adds an edge (u,v) so that in the adjacency list of u, new vertex v is 
        before a, and in the adjacency list  of v, vertex u is after b 
        if a or b are not in the adjacency list, the function
        simply appends u and v to the relevant lists. (So, then it behaves like the standard
        addEdge. (This last behavior is useful if we just start building the graph.)
        '''
        listU = self.adj[u]
        if a in listU:
            i = listU.index(a)
            self.adj[u].insert(i, v)
        else:
            self.adj[u].append(v)
        listV = self.adj[v]
        if b in listV:
            i = listV.index(b)
            self.adj[v].insert((i + 1)%self.degree(v), u) 
        else:
            self.adj[v].append(u)
        
  
    def removeEdge(self, u, v):
        '''
        removes edge (u, v)
        '''
        self.E = self.E - 1  
        #print("list of u: " + str(self.adj[u]))
        self.adj[u].remove(v)
        #print("list of v: " + str(self.adj[v]))
        self.adj[v].remove(u)
       
    def removeVertex(self, v):
        '''
        returns graph in which vertex v and all edges adjacent to it are removed.
        The assumption is 
        that it is not a root or rootArrow
        '''
        vertices = list(range(v)) + list(range(v + 1, self.V))
        gr = self.subgraph(vertices, self.root, self.rootArrow)
        return gr
    
    def addVertex(self): 
        '''
        adds a vertex to the graph
        '''
        self.V = self.V + 1
        self.adj.append([])
        self.repr.append([0., 0.])
        
                       
    
    # Sets coordinates of vertex v to (x, y)    
    def setCoord(self, v, x, y):
        #print("len(repr) = " + str(len(self.repr)))
        #print("v = " + str(v))
        self.repr[v] = [x, y]
    
    '''
    represent vertices in n-by-1 array vertices
    as points with coordinates in n-by-2 array vrepr
    '''
    def reprVertices(self, vertices, vrepr):
        for i in range(len(vertices)):
            self.setCoord(vertices[i], vrepr[i][0],vrepr[i][1])
            
    '''
    represent selected vertices on the unit circle
    '''        
    def reprOnCircle(self, vertices):
        n = len(vertices)
        for k in range(n):
            self.setCoord(vertices[k], math.cos(2 * math.pi * k/n),
                           math.sin(2 * math.pi * k/n)) 
            
               
    def putSingletonsAway(self):
        '''
        represents singletons by points in the south-west corner
        '''
        sngls = self.singletons()
        n = len(sngls)
        vrepr = [[-1,-1]] * n
        self.reprVertices(sngls, vrepr)

    def calcRubberRepr(self, nailedVs):
        '''
        calculates representation of the graph using the 
        current representation of nailed vertices as fixed and using 
        rubber method that minimizes the energy of edges that connect vertices
        
        Reference: Lovasz "Graphs and Geometry"
        '''
        Alist = [];
        for v in range(self.V):
            equation = []
            for i in range(self.V):
                equation.append(0)
            equation[v] = 1
            if v not in nailedVs: #the equations are non-trivial only for these vertices
                d = len(self.adj[v]) #number of neighbors
                for w in range(d):
                    equation[self.adj[v][w]] = - 1/d
            Alist.append(equation) 
        A = np.array(Alist)
        #print("A is ")
        #print(A) 
        #now we build right hand sides and solve for representation
        b = []
        for v in range(self.V):
            if v not in nailedVs:
                b.append(0)
            else:
                b.append(self.repr[v][0])
        bx = np.array(b)
        #print("bx is ")
        #print(bx)
        x = la.solve(A,bx)
        #print(" x is ")
        #print(x)  
        #now solving for y coordinates      
        b = []
        for v in range(self.V):
            if v not in nailedVs:
                b.append(0)
            else:
                b.append(self.repr[v][1])
        by = np.array(b)
        y = la.solve(A,by)
        #print(" y is ")
        #print(x)
        #adjusting the representation of vertices
        for v in range(self.V):
            self.repr[v][0] = x[v]
            self.repr[v][1] = y[v]  
    
    def calcKoebeRepr(self):
        '''
        calculates representation of the graph as a tangency graph for a system of circles
        reference: Lovascz "Graph and Geometry"
        
        Arguments: 
        '''
        if (self.rootArrow < 0):
            print("calcKoebeRepr says:")
            print("the rootArrow element")
            print("of the graph should be non-negative in order to calculate bounndary")
        fixiki = self.getOuterBoundary(u0 = self.rootArrow)
        self.reprOnCircle(fixiki)
        self.calcRubberRepr(fixiki)
        if self.dlgr == None:
            self.dualGraph()
        self.findRadii()
        #caclculate angles
        angles = self.calcAngles() 
            #Angles - the p-th element contains the list of angles for the vertices of the country
            #p. The angle is the half of the inner angle for the vertex. 
        
        
        #reset all vector positions to 0
        #(I should have represented vertices by numpy vectors from the beginning)
        for v in range(self.V):
            self.repr[v][0] = 0.
            self.repr[v][1] = 0.
        v0 = self.countries[0].border[0] #we place this vertex at 0
        v1 = self.countries[0].border[1] #and this vertex will be on the horizontal line.
                                #all other vertices should be determined automatically
        self.repr[v1][0] = self.R[v0] + self.R[v1]
        markedVertices = [False] * self.V
        markedVertices[v0] = True
        markedVertices[v1] = True
        #we are going to traverse the dual graph and build coordinates for vertices of 
        #each country
        markedCountries = [False] * self.dlgr.V
        markedCountries[0] = [True] #We will not process the outer country
        country = self.getCountry(v1, v0)
        stack = [country]
        counter = 0
        while len(stack) > 0 and counter < 2 * (self.V + self.E):
            counter = counter + 1
            #print("Current repr = " + str(self.repr) )
            country = stack[-1]
            p = self.countries.index(country)
            markedCountries[p] = True
            #print("Country = " + str(country) + "; index = " + str(p))
            #process the country: find the coordinates of all its vertices
            x = 0
            border = country.border
            l = len(border)
            for i in range(l):
                u = border[i - 1]
                v = border[i]
                if markedVertices[u] and markedVertices[v]:
                    x = i
                    break #found a pair of vertices with known coordinates
            for i in range(l - 1):
                u = border[(i + x - 1) % l]
                v = border[(i + x) % l]
                w = border[(i + x + 1) % l]
                if not markedVertices[w]: #need to calculate the representation of vertex w
                    #print("calculating representation for w = " + str(w))
                    #print("index of vertex in the country is (i + x)%l = " + str((i + x) % l))
                    #print("angles for country " + str(p) + " are " + str(angles[p]))
                    a = 2 * angles[p][(i + x) % l]
                    #print("Angle next to v = " + str(v) + " is " + str(180 * a/np.pi))
                    uv = np.array([self.repr[u][0] - self.repr[v][0],
                                   self.repr[u][1] - self.repr[v][1]])
                    vrepr = np.array([self.repr[v][0], self.repr[v][1]])
                    #print(vrepr)
                    wv = np.array([uv[0]*np.cos(a) - uv[1]* np.sin(a), 
                                   + uv[0]*np.sin(a) + uv[1]*np.cos(a)])
                    wv = wv/(np.sqrt(wv @ wv))
                    wv = wv * (self.R[w] + self.R[v])
                    wrepr = vrepr + wv
                    self.repr[w][0] = wrepr[0]
                    self.repr[w][1] = wrepr[1] 
                    #print("calculated position for  w = " + str(w) + " is " + str(wrepr))
                    markedVertices[w] = True
            #look for new countries
            foundNewCountry = False
            for i in range(len(country.border)):
                u = country.border[i - 1]
                v = country.border[i]
                newCountry = self.getCountry(v,u)
                newp = self.countries.index(newCountry)
                if not markedCountries[newp]:
                    foundNewCountry = True
                    stack.append(newCountry)
                    #print("newCountry: " + str(newCountry))
                    break
            if not foundNewCountry:    
                stack.pop()
              
    def calcSchnyderRepr(self):
        '''
        calculated the coordinates of Schnyder's representation for a triangulation graph.
        sets the coordinates of vertices placing the vertices of the external triangle
        as vertices of an equilateral triangle on the unit cirle. 
        
        Returns Shnyder's integer barycentric coordinates as an np.array X
        
        Reference: Schnyder "Embedding planar graphs on a grid"
        '''
        
        #seq = self.canonLabelling()
        tree_arrays = self.calcRealizer()
        #now should convert arrays to trees (the trees may have the planar structure different from 
        #what they have in the original graph
        trees = []
        for t in range(3):
            T = tree_arrays[t]
            trees.append(ot.OrderedTree(V = len(T)))
            for i, v in enumerate(T):
                if v != -1:
                    trees[t].graph.addEdge(i, v)
                    if T[v] == -1:
                        #print("found root v = " + str(v))
                        trees[t].setRoot(v)
        P = [] #heights of vertices in the three trees
        Q = [] #sizes of subtrees rooted on vertices in the three trees               
        for t in range(3):                
            heights, sizes = trees[t].calcHeightsSizes(trees[t].root)
            #print("heights of tree " + str(t) + " is " + str(heights))
            #print("sizes of tree " + str(t) + " is " + str(sizes)) 
            P.append(heights)
            Q.append(sizes)      
            
        #now we are going to calculate the number of vertices in the three regions that 
        # correspond to each vertex. We are going to do using the dfs traversals over the trees
        #initializing
        R = []
        for _ in range(3):
            rrr = [0] * self.V
            R.append(rrr)   
           
        for t in range(3):
            tree = trees[t]
            
            cumsum1 = [0] * self.V
            cumsum2 = [0] * self.V #here we keep cumsums of sizes of subtrees
            #A[x] is the order in which x appears in the dfs path
            A = [-1]* self.V
            
            stack = [tree.root]
            A[tree.root] = 0
            
            n = 1
            while (len(stack) > 0):
                #print("stack = " + str(stack));
                v = stack[-1]; #peaking at the top of the stack
                #print("processing v = " + str(v))
                flag = False #can we find unprocessed neighbors?
                d = tree.graph.degree(v)
                for i in range(d):
                    w = tree.graph.adj[v][i]
                    if (A[w] < 0): #this neighbor has not been explored previously
                        A[w] = n
                        n = n + 1
                        stack.append(w)
                        cumsum1[w] = cumsum1[v] + Q[t - 1][w]
                        cumsum2[w] = cumsum2[v] + Q[(t + 1) % 3][w]
                        R[t - 1][w] = R[t - 1][w] + cumsum1[w]
                        R[(t + 1) % 3][w] = R[(t + 1) % 3][w] + cumsum2[w]
                        R[t][w] = R[t][w] - Q[t][w]
                        flag = True
                        break #stop searching for an unexplored neighbor
                    else: 
                        continue
                if (not flag): 
                    stack.pop() #no unexplored neighbor around v. Remove it from consideration.
        #print(R) 
        #finally let us calculate the coordinates of the vertices
        X = np.zeros((3, self.V))
        for t in range(3):
            X[t, :] = np.array(R[t]) - np.array(P[t - 1]) + 1
        #adjust the cooordinates of exterior vertices
        for t in range(3):
            X[t, trees[t].root] = self.V - 2
            X[t - 1, trees[t].root] = 0
            X[(t + 1) % 3, trees[t].root] = 1   
        #now we calculate the representation 
        basis = np.transpose(np.array([[1, 0], [-1/2, np.sqrt(3)/2], [-1/2, - np.sqrt(3)/2]]))
        #print(basis)
        Y = basis @ X
        #print(Y)
        for i in range(self.V):
            self.repr[i][0] = Y[0,i]
            self.repr[i][1] = Y[1,i]
        return X    
  
    def canonLabelling(self):
        '''
        build a canonical labelling for a maximal planar graph (that is, for a triangulation)
        with outer face a, b, c
        That is, outputs a sequence [v0, v1, ..., v_n], where v0 = b, v1 = c, v_n = a, and the 
        sequence satisfies certain requirements useful for building a planar embedding of the
        graph. See Schnyder paper (1990) about realizers and references within. 
        It is assumed that the triangulation has at least 4 vertices 
        '''
        gr = copy.deepcopy(self)
        V = self.V
        seq = [0] * V
          
        boundary = gr.getOuterBoundary(u0 = gr.rootArrow)
        b = boundary[0]
        c = boundary[1]
        a = boundary[2]
        seq[0] = b 
        seq[1] = c
        seq[-1] = a
        
        for i in range(V- 3): #we are going to remove V - 3 non-boundary vertices
            nbrs = gr.adj[a]
            #print("neighbors are " + str(nbrs))
            #we are looking for a vertex v among neighbors of vertex a, 
            #which is not b or c and which 
            #has exactly 2 common neighbors with vertex a
            #In terminology of Schnyder, av is contractible.
            isContractionFound = False
            for v in nbrs:
                if (v == b) or (v == c):
                    continue
                else:
                    count = 0 
                    for u in gr.adj[v]: 
                        if (u in nbrs):
                            count = count + 1
                    if count == 2: #a contractible edge is found
                        isContractionFound = True
                        #print("Contractible edge is (" + str(a) + ", " + str(v) + ")")
                        break 
            if not isContractionFound:
                print("canonLabelling says: Contractible edge is not found")
                print("drawing current graph:")
                gr.draw(drawLabels = True)
                break
            else:
                seq[gr.V - 2 - i] = v
                #contracting the graph
                for u in gr.adj[v]: #first we add edges to a
                    #print(u)
                    u0 = v #this is to keep track of which new edges were added to a
                        #if no edges are added, this is v
                    if (u not in nbrs) and (u != a):
                        #print("Nontrivial vertex u = " + str(u))
                        gr.addEdgeBeforeAfter(u, a, v, u0)
                        u0 = u
                toRemove = copy.deepcopy(gr.adj[v])
                for u in toRemove: #now we remove all edges from v
                    #print("removing " + str(u))
                    gr.removeEdge(v, u)
                #print("After contraction of (" + str(a) + ", " + str(v) + ")")
                #print("The graph is " + str(gr))
                #gr.draw(drawLabels = True)
        return seq 
  
    def calcRealizer(self):
        '''
        Builds a Schnyder realizer of the triangulation graph. 
        Arguments:
        graph - a triangulation
        seq - a canonical labelling build by canonLabelling function
        Output:
        a triple of arrays, each of them gives a map from a vertex of a triangulation to its parent
        in one of 3 trees.
        if we have b root, c rootarrow, and a the remaining vertex on the outer triangle
        then T1 has its root at a, T2 - at b, and T3 - at c.
        [The array representation of trees loses the info about planar embeddings of these trees]
        (TODO: may be it would be useful to find a way to output a triple of ordered trees)
        '''
        seq = self.canonLabelling()
        gr = copy.deepcopy(self)
        V = gr.V
        T1 = [-1] * V
        T2 = [-1] * V
        T3 = [-1] * V
        processed = seq[:2]
        for s in range(2, V - 1):
            x = seq[s]
            wheel = gr.adj[x]
            #print("wheel = " + str(wheel))
            deg = len(wheel)
            for i in range(deg): #we need to find such a point v in the wheel of x that v is in 
                #processed list but the vertex preceding v is not
                if (wheel[i - 1] not in processed) and (wheel[i] in processed):
                    wheel = wheel[i:] + wheel[:i]
                    break
            #print("wheel = " + str(wheel))
            T1[x] = seq[-1] #connect current vertex to a in T1 tree
            T2[x] = wheel[0] 
            for i in range(1, deg - 1):
                if (wheel[i] in processed) and (wheel[i + 1] in processed):
                    T1[wheel[i]] = x
                elif (wheel[i] in processed) and (wheel[i + 1] not in processed):
                    T3[x] = wheel[i]
                else:
                    break
            processed.append(x)
        return [T1, T2, T3]   
    
    def degreeProfile(self):
        '''
        Calculates an array with the numbers of vertices that have a given degree.
        Leaves here have degree 1. Singletons are not allowed.
        So [10, 3, 4, 1] means that 10 vertices are leaves, 3 vertices have degree 2, and so on
        The maximum degree is 4
        The array truncated, so that any trailing 0 are removed. 
        '''
        profile = [0] * (self.V - 1)
        maxdegree = 0; 
        for v in range(self.V):
            d = len(self.adj[v])
            if d > maxdegree:
                maxdegree = d
            profile[d-1] = profile[d-1] + 1
        profile = profile[:maxdegree]
        return profile    
    
    def maxDegree(self):
        '''
        calculates the max degree of the graph. 
        '''
        md = 0
        for v in range(self.V):
            d = len(self.adj[v])
            if d > md:
                md = d 
        return md
                
    def subgraph(self, vertices, root, rootArrow):
        '''
        return the graph which is the subraph of this graph based on a list of vertices
        The vertices are renamed to 0, 1, 2, 3, ....
        '''   
        newV = len(vertices)
        sgr = PlanarGraph(newV, root = vertices.index(root), 
                          rootArrow = vertices.index(rootArrow))
        for i in range(newV):
            v = vertices[i]
            sgr.repr[i] = self.repr[v]
            for u in self.adj[v]:
                if u in vertices:
                    sgr.addDirectedEdge(i, vertices.index(u))
        return sgr    
    

    def dualGraph(self):
        '''
        builds the dual graph
        '''
        F = self.E + 2 - self.V #this includes the outer country
        self.dlgr = PlanarGraph(F)
        #we start the list of all countries with the outer country
        outerCountry = Country(self.getOuterBoundary(u0 = self.rootArrow))
        self.countries = [outerCountry]
        #next we should do something like depth first exploration 
        #of the countries, however, we need to check if we have already visited a country
        stack=[outerCountry]
        while (len(stack) > 0):
            currentC = stack[-1] #last element in the stack
            #now we are going over vertices in the current country and 
            #see if we can add any of the neighbor countries
            #print("Current country = " + str(currentC))
            foundNewCountry = False
            for i in range(len(currentC.border)): 
                #get 2 consequitive points 
                u = currentC.border[i - 1]
                v = currentC.border[i]
                neighbor = self.getCountry(v, u)
                #print("The country next to [" + str(v) + ", " + str(u) + "] edge is " 
                #      + str(neighbor))
                if neighbor not in self.countries: #new country is found.
                    foundNewCountry = True
                    self.countries.append(neighbor)
                    stack.append(neighbor)
                    break #stop searching for neighbors of currentC
            if not foundNewCountry:
                stack.pop() #remove current country from the stack
        #print("Countries = " + str(self.countries))
        #add edges to the dual graph
        for x in range(len(self.countries)):
            country = self.countries[x]
            for i in range(len(country.border)):
                u = country.border[i - 1]
                v = country.border[i]
                neighbor = self.getCountry(v, u)
                y = self.countries.index(neighbor) 
                self.dlgr.addDirectedEdge(x, y)
                #print("Neighbor of country " + str(x) + " is " + str(y))
        #print(" Dual Graph is " + str(self.dlgr))
        # assign vector representations to vertices of the dual graph
        for x in range(len(self.countries)):
            country = self.countries[x]
            for v in country.border:
                self.dlgr.repr[x][0] = self.dlgr.repr[x][0] + self.repr[v][0]
                self.dlgr.repr[x][1] = self.dlgr.repr[x][1] + self.repr[v][1]
            self.dlgr.repr[x][0] = self.dlgr.repr[x][0] / len(country.border)
            self.dlgr.repr[x][1] = self.dlgr.repr[x][1] / len(country.border)
                   
    def calcInitAngles(self):
        '''
        calculate original angles: B_{ip} is half of the angle of country p at 
        its vertex i. 
        We exclude the outer country p
        The assumption is that the rubber representation is already computed.
        '''
        #initialize
        V = self.V
        F = self.dlgr.V
        B = [] 
        for _ in range(V):
            B.append([0.0] * (F - 1))
        B = np.array(B)
        #going along countries (except outer Counter)
        for p in range(1, F):
            country = self.countries[p]
            #print(self.countries[p])
            L = len(country.border)
            for i in range(L):
                u = np.array(self.repr[country.border[i - 1]])
                v = np.array(self.repr[country.border[i]])
                w = np.array(self.repr[country.border[(i + 1) % L]])
                x = w - v
                y = u - v
                angle = np.arccos(x @ y /np.sqrt(x @ x * y @ y))/2
                B[country.border[i]][p - 1] = angle
            
        #print(str(B))
        #do a couple of checks
        #Sum over row should give pi or pi/6
        #print(str(np.sum(B, axis = 1)))
        return B
        
    def targetFunction(self, x, B):
        '''
        calculates target function, which will be optimize in order to 
        find the radii of the circles in the circle representation of the planar graph.
        For reference, see Lovasz "Graphs and Geometry" section 5.1.3.
        Arguments:
        x is the list of radii. The first V elements of the list are radii of the circles 
        around vertices, and the next (F - 1) elements are radii of the circles around
        the centers of the countries.
        B is the matrix of the angles calculated by calcInitAngles
        '''
        #let us normalize x so that its maximal element is not greater than 1
        m = np.max(x)
        x = x - m
        V = self.V
        F = self.dlgr.V
        xV = x[:V]
        xP = x[V:]
        phi = 0.0
        for i in range(V):
            for p in range(F - 1):
                if i in self.countries[p + 1].border: #if country and vertex are incident
                    phi = phi + tu.verdiereFunction(- xV[i] + xP[p]) - B[i][p]*(- xV[i] + xP[p])
        return phi
    
    def tfGrad(self, x, B):
        '''
        calculates the gradient of the target function
        '''
        m = np.max(x)
        x = x - m
        V = self.V
        F = self.dlgr.V
        xV = x[:V]
        xP = x[V:]
        grad = [0.0] * len(x)
        for i in range(V):
            for p in range(F - 1):
                if i in self.countries[p + 1].border: #if country and vertex are incident
                    grad[i] = grad[i] - np.arctan(np.exp(- xV[i] + xP[p])) + B[i][p]
                    grad[V + p] = grad[V + p] + np.arctan(np.exp( - xV[i] + xP[p])) - B[i][p] 
        return np.array(grad)
    
    def tfGradient(self, B):
        '''
        this is for checking. It calculates the gradient of the target function in numerical
        fashion -- namely using automatic differentiation
        '''
        fun = lambda x: self.targetFunction(x, B)
        dfun = nd.Gradient(fun)
        return dfun
        
     
    def findRadii(self, useGrad = True, disp = False):   
        '''
        finds the radii of the circles in the representation of the graph
        as the tangency graph for a collection of circles (Koebe's representation).
        This is done by minimization of the targetFunction
        The radii are stored to self.R variable
        
        [see Lovasz "Graphs and Geometry" section 5.1.3.]
        
        If useGrad = True, uses the analytical gradient in tfGrad, otherwise uses
        numerical gradient for optimization
        
        If disp = True, shows the messages generated by optimization routine
        
        '''
        x0 = np.array([1.] * (self.V + self.dlgr.V - 1)) 
        B = self.calcInitAngles();
        if useGrad:
            logR = minimize(self.targetFunction, x0, args = B, method='BFGS', jac=self.tfGrad,
                       options={'disp': disp, 'gtol': 1e-08, 'maxiter': 800})
        else:
            logR = minimize(self.targetFunction, x0, args = B, method='BFGS',
                       options={'disp': disp, 'gtol': 1e-06, 'maxiter': 800})
        print("findRadii says: Target function after optimization = " + str(logR['fun']))
        #print("logR = " + str(logR['x']))
        x = logR['x']
        x = x - np.max(x)
        self.R = np.exp(x)
        
    
    def calcAngles(self):
        '''
        calculate the angles in the circles representation of the graph
        based on the radii computed in findRadii
        The output is an array of list for each country, it 
        corresponds to the ordering of vertices in self.country structure
        '''
        if len(self.R) == 0:
            self.findRadii()
        countryAngles = []
        for p in range(len(self.countries)):
            if p == 0:
                countryAngles.append([0.0] * len(self.countries[0].border))
            else:
                country = self.countries[p].border
                x = [0.0] * len(country)
                for i in range(len(country)):
                    x[i] = np.arctan(self.R[self.V + p - 1]/self.R[country[i]])
                countryAngles.append(x)
        return countryAngles
            
    def calcVertexAngles(self, countryAngles):   
        '''
        calculates the angles for every vertex
        Arguments: the list of angles for countries
        Returns: an array of lists, each least contain angles for a vertex
        The angle for the infinite country is not listed.
        '''     
        vertexAngles = []
        for v in range(self.V):
            anglesV = []
            for w in self.adj[v]:
                p = self.countries.index(self.getCountry(w, v))
                #print("a country for vertex " + str(v) + " is " + str(p))
                if p != 0:
                    i = self.countries[p].border.index(v)
                    anglesV.append(countryAngles[p][i])
            vertexAngles.append(anglesV)
            #print("current vertexAngles: " + str(vertexAngles))
        return vertexAngles      
    
    def drawNiceDualGraph(self):  
        '''
        this simply gives a nice picture of the dual graph. 
        The original graph should be a planar triangulation
        '''
        if self.dlgr == None:
            self.dualGRaph()
        dual = deepcopy(self.dlgr)
        a = self.dlgr.adj[0][0]
        b = self.dlgr.adj[a][0]
        dual.root = b
        dual.rootArrow = a
        dlgr2 = dual.removeVertex(0)
        fig, ax = dlgr2.draw(drawLabels = False, calculateRepr = "drawCircles")
                

    def draw(self, title = '', drawLabels = False, block = False, root = -1, 
             calculateRepr = "Boundary", boundary = [], showDual = False,
             palette = "prism", ax = None):
        '''
        draws the graph.
        Arguments 
        - title = string in the title of the plot
        - drawlabels = if true show labels of vertices next to them. If no Labels are provided
            just show the order number of vertices
        - block: if block is true, then calling 
            the draw function will stop execution of further commands. 
            if block is false (default), then the execution continues but the calling function
            should have plt.show() at the end. 
        - root: if root>0 show this vertex as root. 
        - calculateRepr: If "No" the function will use 
            existing vector representation of vertices;
                if calculateRepr = "Leaves", calculates the vector representation of vertices
            by putting leaves on the unit circle and using rubber edges algorithm
                if calculateRepr = "Boundary" (default), then calculates the vector representation of vertices
            by putting the boundary vertices on the circle. If the boundary argument is non-empty, 
            then it is used as the boundary. Otherwise, the boundary is computed, and 
            then it required that the graph should have positive rootArrow. The assumption 
            is that the edge (root, rootArrow) starts a walk on the boundary. 
                if drawCircles, draws the circles representation.
                [TODO] if calculateRepr = "Schnyder", calculates the Schnyder representation of 
                the graph. This is currently restricted only to triangulations. 
        - boundary: If available then vertices in the boundary are put on the unit circle.
        - showDual: If true, shows the dual graph in red.
                - drawCircles: If true, draws the circles associated with vertices. In this case, 
                 the attribute R should be already computed for the graph (by findRadii function). 
        - palette: the palette used to draw circles, default = "prism"
        
        returns fig and ax handles
         '''
        if (calculateRepr == "Leaves"):  
            boundary = self.getOuterBoundary(typ = "Leaves")
            fixiki = []
            for v in boundary:
                if self.degree(v) == 1:
                    fixiki.append(v)
            #fixiki = self.leaves()
            self.reprOnCircle(fixiki)
            self.calcRubberRepr(fixiki)
            self.putSingletonsAway() 
        elif (calculateRepr == "Boundary"):
            if len(boundary) == 0:
                if (self.rootArrow < 0):
                    print("Draw says: if calculateRepr option is Boundary, ")
                    print("then either the boundary should be supplied or rootArrow element")
                    print("of the graph should be non-negative")
                    return
                else:
                    boundary = self.getOuterBoundary(u0 = self.rootArrow)
                    print("Draw says: Calculated Boundary = " + str(boundary))
            fixiki = boundary.copy()
            self.reprOnCircle(fixiki)
            self.calcRubberRepr(fixiki)
            self.putSingletonsAway()
        elif calculateRepr == "drawCircles":
            print("calculating Koebe's representation")
            self.calcKoebeRepr()
            #print(self.R)
        elif calculateRepr == "Schnyder":
            X = self.calcSchnyderRepr()
            
        
        if (ax == None):
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
            
        ax.set_title(title)
        #TODO choice of plot limits can be improved
        A = np.array(self.repr)
        x = A[:, 0]
        y = A[:, 1]
        ax.set_xlim(np.amin(x), np.amax(x))
        ax.set_ylim(np.amin(y), np.amax(y))
        ax.set_xticks([])
        ax.set_yticks([])

        for v in range(self.V):
            for w in self.adj[v]:
                if w > v: #to avoid double-drawing of the same edge
                    line = mlines.Line2D(np.vstack((x[v], x[w])),
                                         np.vstack((y[v], y[w])), marker = ".");
                    ax.add_line(line)
                    
        if root >= 0:
            circle = plt.Circle((x[root], y[root]), 0.01, color='g', clip_on=False)
            ax.add_artist(circle)
            ax.scatter(x[root], y[root], marker = "X", color = 'g')
        else:
            circle = plt.Circle((x[self.root], y[self.root]), 0.01, color='g', clip_on=False)
            ax.add_artist(circle)
            
        if drawLabels:
            if len(self.labels) == 0: #use index of the vertex instead of label
                for v in range(self.V):
                    ax.text(x[v],y[v], str(v))
            else:
                for v in range(self.V):
                    ax.text(x[v],y[v], str(self.labels[v]))
                    
        if showDual:
            A = np.array(self.dlgr.repr)
            x = A[:, 0]
            y = A[:, 1]
            for v in range(1, self.dlgr.V):
                if drawLabels:
                    ax.text(x[v],y[v], str(v))
                for w in self.dlgr.adj[v]:
                    if w > v:
                        line = mlines.Line2D(np.vstack((x[v], x[w])),
                            np.vstack((y[v], y[w])), marker = ".", color = "r");
                    ax.add_line(line)
                    
        if calculateRepr == "drawCircles":
            colormap = cm.get_cmap(palette, 128)
            #print(self.R)
            maxLogR = np.max(-np.log(self.R))
            for v in range(self.V):
                c = colormap(-np.log(self.R[v])/maxLogR)
                circle = plt.Circle((self.repr[v][0], self.repr[v][1]),
                    self.R[v], color=c, clip_on=False)
                ax.add_artist(circle)
        
        if block:            
            plt.show()
        else:
            plt.draw()
        return fig, ax
    
    def drawPath(self, ax, path):
        '''
        draws a path that connects vertices in path argument in the axes given by ax
        The axes should be created by draw function
        '''    
        A = np.array(self.repr)
        x = A[:, 0]
        y = A[:, 1]       
        for i in range(len(path) - 1):
            v = path[i]
            w = path[i + 1]
            line = mlines.Line2D(np.vstack((x[v], x[w])),
                                         np.vstack((y[v], y[w])), color = 'r', 
                                         linestyle = '-');
            ax.add_line(line)
    
    #def drawRealizer(self, ax, realizer):
        '''
        draws graph with a Schnyder's realizer (which was computed by calcRealizer)
        '''
    '''
        A = np.array(self.repr)
        x = A[:, 0]
        y = A[:, 1]
        for i in range(3):
            if i == 0:
                c = "red"
            elif i == 1:
                c = "green"
            else:
                c = "purple"
            T = realizer[i]
            for i, v in enumerate(T):
                if v >= 0:
                    line = mlines.Line2D(np.vstack((x[i], x[v])),
                                     np.vstack((y[i], y[v])), marker = ".", color = c);
                    ax.add_line(line)
                    arrow = mpatches.Arrow(x[i], y[i], x[v] - x[i], y[v] - y[i], width = 0.1,
                                            color = c);
                    ax.add_patch(arrow)
    '''


    
       
    def outerFaceClosure(self):
        '''
        goes along the edges in the outer face starting from the root and 
        in the counterclockwise direction. Adds a triangular face if possible.
        This function will only work if certain assumptions are satisfied 
     
        Returns 1 if a face was added or 0 if it was not added
        '''  
        #first three edges from the root along the boundary of the outer space.
        x = self.root   #note that we almost always use (or should use)
                        #the dfs names for vertices
                        #and in this convention the root is always named 0.
        y = self.adj[x][0]
        z = self.nextEdge(y, x)
        t = self.nextEdge(z, y)
        #print("(x, y, z, t) = (" + str([x,y,z,t]) + ")")
        counter = 0
        while ((t == y or x == z or (self.degree(x) == 1 and self.degree(t) == 1) 
               or (self.degree(x) > 1 and self.degree(t) > 1))
               and counter < 2 * self.V): #go along the boundary until find a suitable sequence 
                                    #of edges or until the number of observed edge sequences >= 2 V
            counter = counter + 1
            #print("x = " + str(x) + "; counter = " + str(counter))
            x = y
            y = z
            z = t 
            t = self.nextEdge(z, y)
        #print("degree of " + str(x) + " = " + str(self.degree(x)))
        #print("degree of " + str(t) + " = " + str(self.degree(t)))
        #print("[x, y, z, t] = "    + str([x,y,z,t]) + ")")
        
        #Now we need to check and if one vertex on the boundary of the sequence is a leaf
        #and  the other is not a leaf, and there is no other leaf in between) then we need to remove the leaf and connect 
        #the parent of the leaf with the other end-point. This connection should be done
        #accurately, so that planarity of the graph would not suffer.
        if (self.degree(x) == 1 and self.degree(t) > 1 and self.degree(z) > 1):
            self.removeEdge(y, x)
            self.addEdgeBeforeAfter(y, t, z, z)
            return 1
        elif (self.degree(x) > 1 and self.degree(y) > 1 and self.degree(t) == 1):
            self.removeEdge(z, t)
            self.addEdgeBeforeAfter(x, z, y, y)
            return 1
        else: 
            return 0
    
    def getCountry(self, u0, v0):
        '''
        gets a country with an edge (u0, v0) on its border so that if we go
        along (u0, v0) the country is on the right. (And so we go in the clockwise direction
        for inner countries.)
        '''
        border = [u0]
        u = u0 #current vertex
        v = v0 #next vertex
        while (v != u0):
            border.append(v)
            w = u #new previous vertex
            u = v #new current vertex
            v = self.nextEdge(u, w)
        country = Country(border)
        if country in self.countries: #this strange thing ensures, that if the country is already
                            #in the list of countries, then its border is exactly as its
                            # in the listed country 
            p = self.countries.index(country)
            return(self.countries[p])
        else: 
            return country   
    
    def getOuterBoundary(self, typ = "root", u0 = -1):
        '''
        gets a sequence of vertices that are on the border of the outer face
        if typ = "root" assumes that root is on the boundary and uses (root, u0) 
        as the first edge in the walk around the boundary.
        Otherwise tries to start from a leaf. 
        It is assumed that all leaves look into the outer face.
        '''
        boundary = []
        if typ == "root": #use the supplied arguments
            if u0 < 0:
                print("getOuterBoundaryWarning: A positive u0 must be supplied")
                return
            else: 
                v0 = self.root
                v = v0 #current vertex
                u = u0 #next vertex
        else: #try to start boundary at a leaf
            lvs = self.leaves()
            if (len(lvs) == 0): #there are no leaves, cannot start
                return boundary
            v0 = lvs[0] #initial vertex
            v = v0 #current vertex on the boundary
            u = self.adj[v0][0] #next vertex in on the boundary
        
        boundary = [v0]
        while (u != v0):
            boundary.append(u)
            w = v #new previous vertex
            v = u #new current vertex
            u = self.nextEdge(v, w)
            if u == -1:
                print("getOuterBoundary says: was not able to continue building boundary")
                print("current boundary is " + str(boundary))
                print("the graph is " + str(self))
                boundary = -1
                return boundary
        return boundary    

class Country(object):
    '''
    An object that represents a country in a planar graph
    '''


    def __init__(self, border):
        '''
        Constructor:
        boundary is the list of vertices that surrounds the country in counterclockwise order
        '''
        self.border = border   
        
    def __str__(self):
        '''
        returns the string representation of the country_names
        '''
        return str(self.border)
    def __repr__(self):
        '''
        returns the string representation of the country_names
        '''
        return str(self.border)
    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Country):
            if self.border == other.border: #borders are same
                return True
            else: #check if borders would coincide after a cyclic shift
                b = self.border.copy()      
                for _ in range(len(self.border) - 1):
                    b = b[1:] + b[:1]
                    #print(b)
                    if b == other.border:
                        return True
        return False
    
def getFromTree(tree):
    '''
    returns a copy of the tree's graph
    '''
    gr = deepcopy(tree.graph)
    fixiki = gr.leaves() 
    gr.reprOnCircle(fixiki)
    gr.calcRubberRepr(fixiki)
    return gr
        
def buildTriangulation(tree):
    '''
    builds a triangulation from the supplied tree
    the tree should be B - tree (every inner vertex has 2 leaves)
    '''
    gr = getFromTree(tree)
    flag = 1
    counter = 0;
    while (flag == 1 and counter < 2 * gr.V):
        #print(counter)
        counter = counter + 1
        flag = gr.outerFaceClosure() 
        #gr.draw(drawLabels = True, block = False)
    if counter == 2 * gr.V :
        print ("buildTriangulation warning: there was too many face closures.")
        
    '''
    In the second step we add two additional vertices and glue leaves to these vertices
    '''
    boundary = gr.getOuterBoundary(typ = "Leaves")
    #print("Boundary = " + str(boundary))

    if (len(boundary) == 0): #this condition is satisfied only if there is no leaves in the 
                            #boundary. In this case the graph must already be a triangulation
        return gr
    extrema =[] #points that occur 3 times in the boundary sequence, and so adjacent to 2 leaves
    for v in range(gr.V):
        if (boundary.count(v) == 3):
            extrema.append(v)
    #print("Extrema = " + str(extrema))
    ind_extrema0 = [i for i in range(len(boundary)) if boundary[i] == extrema[0]]
    #print(ind_extrema0)
    
    #let us rotate the boundary so that that it starts with the first extreme
    #and this extreme is between two leaves
    #We well do it by checking three possibilities for indices of extrema[0]
    flag = False
    for s in range(3):
        boundary_rot = boundary[ind_extrema0[s]:] + boundary[:ind_extrema0[s]]
        #print("Rotated boundary = " + str(boundary_rot))
        j = boundary_rot.index(extrema[1])
        #print(j)
        lvs1 = []
        for k in range(1, j + 2):
            v = boundary_rot[k]
            if gr.degree(v) == 1:
                lvs1.append(v)
                lvs2 = []
        for k in range(j + 3, len(boundary_rot)):
            v = boundary_rot[k]
            if gr.degree(v) == 1:
                lvs2.append(v)
        #check that this lvs1 has only one leaf connected to extrema0
        count = 0
        for leaf in lvs1:
            if gr.adj[leaf][0] == extrema[0]:
                count = count + 1
        if count == 1: #found a suitable sequence of leaves
            flag = True
            break 
    if not flag:
        print("buildTriangulation says: is not able to find a valid lvs1 sequence of leaves")
        print("Graph before gluing leaves to additional vertices: " + str(gr))
        print("lvs1 = " + str(lvs1))
        print("lvs2 = " + str(lvs2))
        gr.draw(drawLabels = True,
            calculateRepr = "Leaves",
            title = "Graph before gluing leaves to additional vertices: ") 
    #print("Graph before gluing leaves to additional vertices: " + str(gr))
    #print("lvs1 = " + str(lvs1))
    #print("lvs2 = " + str(lvs2))
              
    #gr.draw(drawLabels = True,
    #        calculateRepr = "Leaves",
    #        title = "Graph before gluing leaves to additional vertices: ")
    #plt.show()

    
    #now we introduce two new vertices - v1 and v2 and glue leaves from lvs1 and lvs2 to them
    gr.V = gr.V + 2
    gr.adj.append([])
    gr.adj.append([])
    gr.repr.append([0., 0.])
    gr.repr.append([0., 0.])
    v1 = gr.V - 2
    v2 = gr.V - 1
    
    for u in lvs1:
        w = gr.adj[u][0]
        if len(gr.adj[v1]) == 0:
            gr.addEdgeBeforeAfter(v1,w, -1, u)
        else:
            gr.addEdgeBeforeAfter(v1,w, gr.adj[v1][0], u)
        gr.removeEdge(w,u)
    #print("Extrema:" + str(extrema))
    gr.root = extrema[0]    
    gr.rootArrow = v1
    #gr.draw(drawLabels = True, block = False,
    #        calculateRepr = "Boundary", title = "Graph after gluing lvs1: " + str(lvs1))
    #print("Graph after gluing lvs1: " + str(gr))
    
    for u in lvs2:
        w = gr.adj[u][0]
        if len(gr.adj[v2]) == 0:
            gr.addEdgeBeforeAfter(v2, w, -1, u)
        else:
            gr.addEdgeBeforeAfter(v2, w, gr.adj[v2][0], u)
        gr.removeEdge(w,u)
        #print("graph after gluing " + str(v2) + " instead of  " + str(u)  + ": " + str(gr))
    #gr.draw(drawLabels = True, block = False,
    #        calculateRepr = "Boundary", title = "Graph after gluing lvs2: " + str(lvs2))
    #print("Graph after gluing lvs2: " + str(gr))
    
    #connect two additional vertices 
    gr.addEdge(v1, v2)
    #gr.root = extrema[0]
    #gr.rootArrow = v2
    #gr.draw(drawLabels = True, block = False,
    #        calculateRepr = "Boundary", title = "Graph before renaming vertices")
    #gr.draw(drawLabels = True, title = "Graph before renaming vertices")
    
    #remove all singletons and rename the remaining vertices
    vvv = []
    for v in range(gr.V):
        if gr.degree(v) > 0:
            vvv.append(v)
    sgr = gr.subgraph(vvv, gr.root, gr.rootArrow)
    return sgr
    

'''
For testing methods
'''
def main():
    print('Graph.main() started')
    '''
    V = 4
    my_graph = PlanarGraph(V)     
    my_graph.addEdge(0,1)
    my_graph.addEdge(1,3)
    my_graph.addEdge(3,0)
    my_graph.addEdge(3,2)
    my_graph.addEdge(2,0)
    my_graph.addEdge(1,2)
    my_graph.root = 1
    vertices = [0, 1, 2, 3]
    vrepr = np.array(([[0, 0, 1, 1], [0, 1, 1, 0]]))
    vrepr = np.transpose(vrepr)
    my_graph.reprVertices(vertices,vrepr)
    print(my_graph.repr)
    print(my_graph)
    nailedVs = [0, 1, 2]
    my_graph.reprOnCircle(nailedVs)
    my_graph.calcRubberRepr(nailedVs)
    my_graph.draw("Graph 1",  block = False, drawLabels = True)
    '''
    
    '''
    some nice pictures
    '''
    '''
    n = 50
    #seed = 123
    seed = 0 #random, that is, changes from time to time  
    gr = randomTriangulation(n, seed)
    print("Triangulation graph = " + str(gr))
    gr.draw(drawLabels = False, block = False, 
           calculateRepr = "Boundary")
    
    gr.dualGraph()   
    gr.draw(drawLabels = False, block = False, 
            calculateRepr = "Boundary", showDual = True)
   '''
    
    '''
    drawing a tangency circle graphs (the dual graph is very pretty)
    '''  
    n = 50
    #seed = 123
    seed = 0 #random, that is, changes from time to time  
    trngl, _ = tr.randomTriangulation(n, seed)
    _, ax = trngl.graph.draw(drawLabels = False, calculateRepr = "drawCircles")
    trngl.graph.dualGraph()  
    trngl.graph.drawNiceDualGraph()
    
    plt.show()

    
    print('Graph.main() finished')  

    '''  
    (It seems that there is a memory leak somewhere. If the program is left 
    for a long time with plots open, then
    it eats all the computer memory. It may be a problem of matplotlib.)
    '''
    
if __name__ == "__main__":
    main()




'''
Created on Jul 11, 2020

@author: vladislavkargin
'''

import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
from matplotlib import cm

import copy
import math
import numpy as np

import rgs.pmaps.OrderedTree as ot
import rgs.pmaps.TreeUtility as tu
import rgs.pmaps.PlanarGraph as pg


class Triangulation(object):
    '''
    An object that represent a triangulation. Its an specialization of the planar graph
    so many of its functionality is provided by the PlanarGraph object.
    '''


    def __init__(self, tree = None):
        '''
        Constructor
        If tree is available buils the triangulation from the tree 
        using a variant of the Poulalhon-Scheffer method, 
        else builds the triangulation that consists of 3 vertices.
        '''
        if tree != None:
            self.graph = pg.buildTriangulation(tree)
            self.realizer = self.calcRealizer()
        else:
            self.graph = pg.PlanarGraph(V = 3, root = 0)
            self.graph.rootArrow = 1
            self.graph.addEdge(0,1)
            self.graph.addEdge(1,2)
            self.graph.addEdge(0,2)
            self.realizer = ([-1, -1, -1], [-1, -1, -1], [-1, -1, -1]) 
    
    def __str__(self):
        return str(self.graph)
    
            
    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Triangulation):
            if self.realizer == other.realizer: #realizers are same
                return True
            else: 
                return False
        return False
    
    def __hash__(self):
        return hash(str(self.realizer))
    
    
    def canonLabelling(self):
        '''
        build a canonical labelling for a maximal planar graph (that is, for a triangulation)
        with outer face a, b, c
        That is, outputs a sequence [v0, v1, ..., v_n], where v0 = b, v1 = c, v_n = a, and the 
        sequence satisfies certain requirements useful for building a planar embedding of the
        graph. See Schnyder paper (1990) about realizers and references within. 
        It is assumed that the triangulation has at least 4 vertices 
        TODO: I should enhance this by renaming so that the root was 0, arrowRoot 1 and 
        the remaining vertex on the boundary 2. Perhaps it should be done in a separate 
        method. 
        '''
        gr = copy.deepcopy(self.graph)
        V = gr.V
        seq = [0] * V
          
        boundary = gr.getOuterBoundary(u0 = gr.rootArrow) #this sometimes fails I don't know
                                                        #why
        if boundary == -1:
            return -1                                              
        b = boundary[0]
        c = boundary[1]
        a = boundary[2]
        seq[0] = b 
        seq[1] = c
        seq[-1] = a
        
        for i in range(V- 3): #we are going to remove V - 3 non-boundary vertices
            nbrs = gr.adj[a]
            #print("neighbors are " + str(nbrs))
            #we are looking for a vertex v among neighbors of vertex a, 
            #which is not b or c and which 
            #has exactly 2 common neighbors with vertex a
            #In terminology of Schnyder, av is contractible.
            isContractionFound = False
            for v in nbrs:
                if (v == b) or (v == c):
                    continue
                else:
                    count = 0 
                    for u in gr.adj[v]: 
                        if (u in nbrs):
                            count = count + 1
                    if count == 2: #a contractible edge is found
                        isContractionFound = True
                        #print("Contractible edge is (" + str(a) + ", " + str(v) + ")")
                        break 
            if not isContractionFound:
                print("canonLabelling says: Contractible edge is not found")
                print("drawing current graph:")
                gr.draw(drawLabels = True)
                break
            else:
                seq[gr.V - 2 - i] = v
                #contracting the graph
                for u in gr.adj[v]: #first we add edges to a
                    #print(u)
                    u0 = v #this is to keep track of which new edges were added to a
                        #if no edges are added, this is v
                    if (u not in nbrs) and (u != a):
                        #print("Nontrivial vertex u = " + str(u))
                        gr.addEdgeBeforeAfter(u, a, v, u0)
                        u0 = u
                toRemove = copy.deepcopy(gr.adj[v])
                for u in toRemove: #now we remove all edges from v
                    #print("removing " + str(u))
                    gr.removeEdge(v, u)
                #print("After contraction of (" + str(a) + ", " + str(v) + ")")
                #print("The graph is " + str(gr))
                #gr.draw(drawLabels = True)
        return seq 
    
    def calcRealizer(self):
        '''
        Builds the Schnyder realizer of the triangulation graph. 
        Arguments:
        graph - a triangulation
        seq - a canonical labelling build by canonLabelling function
        Output:
        a triple of arrays, each of them gives a map from a vertex of a triangulation to its parent
        in one of 3 trees.
        if we have b root, c rootarrow, and a the remaining vertex on the outer triangle
        then T1 has its root at a, T2 - at b, and T3 - at c.
        [The array representation of trees loses some info about planar embeddings of these trees]
        '''
        return self.graph.calcRealizer()
        
            
    def draw(self, title = '', drawLabels = False, block = False, root = -1, 
             calculateRepr = "Schnyder", boundary = [], showDual = False,
             palette = "prism"):        
        fig, ax = self.graph.draw(title = title, drawLabels = drawLabels, block = block, root = root, 
             calculateRepr = calculateRepr, boundary = boundary, showDual = showDual,
             palette = palette)   
        return fig, ax
    
    def drawRealizer(self, ax, realizer):
        '''
        draws graph with a Schnyder's realizer (which was computed by calcRealizer)
        '''
        #fig, ax = self.draw(calculateRepr = calculateRepr)
        A = np.array(self.graph.repr)
        x = A[:, 0]
        y = A[:, 1]
        for i in range(3):
            if i == 0:
                c = "red"
            elif i == 1:
                c = "green"
            else:
                c = "purple"
            T = realizer[i]
            for i, v in enumerate(T):
                if v >= 0:
                    line = mlines.Line2D(np.vstack((x[i], x[v])),
                                     np.vstack((y[i], y[v])), marker = ".", color = c);
                    ax.add_line(line)
                    arrow = mpatches.Arrow(x[i], y[i], x[v] - x[i], y[v] - y[i], width = 0.1,
                                            color = c);
                    ax.add_patch(arrow)

def randomTriangulation(n, SEED = 0, method = "Lukas31"):
    '''
    builds a random triangulation on n vertices
    arguments:
    n - number of vertices in the triangulation
    SEED is the seed of the random generator for replication;
        if SEED = 0 (default) then the results will change from time to time.
    method - if "Lukas31" (default) uses the Poulalhon - Schaeffer method
        to build a uniformly random triangulation based on building a Lukasiewisz path
        with steps 3 and -1 and then decoding it to a tree, in which each internal vertex
        is connected to two leaves.
        -if "decoratedPath" then modifies the Poulalhon Schaeffer method and builds 
        a tree in which an internal vertex is connected to two leaves and exactly one next
        internal vertex, so it is a path decorated with leaves (or some kind of asymmetric
        caterpillar. 
    returns the triangulation and the tree from which it was constructed
    '''
    if method == "Lukas31":
        tree = ot.randomBTree(n - 2, SEED = SEED)
        #tree.draw(drawLabels = True)
    elif method == "decoratedPath":
        tree = tu.decoratedPathTree(n, SEED = SEED)
    else:
        print("randomTriangulation says: the method should be either Lukas31 or decoratedPath")
    trngl = Triangulation(tree)
    return trngl, tree 


'''
For testing methods
'''
def main():
    print('Triangulation.main() started')
    
    

    
    #n = 8
    #seed = 23167
    '''
    n = 5
    seed = 0
    trngl, _ = randomTriangulation(n, SEED = seed)
    #trngl = randomTriangulation(n, SEED = seed, method = 'decoratedPath')
    print(trngl.graph)
    _, ax = trngl.draw(drawLabels = True)
    

    profile = trngl.graph.degreeProfile()
    print("Degree profile is " + str(profile))
    print("Max degree = " + str(trngl.graph.maxDegree()))
    boundary = trngl.graph.getOuterBoundary(u0 = trngl.graph.rootArrow)
    print("Calculated Boundary = " + str(boundary))
    seq = trngl.graph.canonLabelling()
    print("Canonical Labeling = " + str(seq))
    
   # realizer = trngl.graph.calcRealizer()
    #print(trngl.realizer) 
    
    X = trngl.graph.calcSchnyderRepr()
    print("barycentric coordinates are: \n" + str(X))   
    
    #_, ax = trngl.draw(drawLabels = True)
    #trngl.drawRealizer(ax, trngl.realizer)
    '''
    
    
    n = 5
    seed = 0
    hashTrngl = {}
    ITER = 100
    for count in range(ITER):
        if count % 10 == 0:
            print(count)
        trngl, _ = randomTriangulation(n, SEED = seed)
        #print(str(trngl))
        seq = trngl.canonLabelling()
        if seq == -1:
            break
        if trngl in hashTrngl.keys():
            hashTrngl[trngl] = hashTrngl[trngl] + 1
        else:
            hashTrngl[trngl] = 1
    print(len(hashTrngl))
    for i in range(len(hashTrngl)):
        trngl = list(hashTrngl.keys())[i]
        _, ax = trngl.draw(drawLabels = True)
        trngl.drawRealizer(ax, trngl.realizer)
        

    
    plt.show()  
    
    print('Triangulation.main() finished')  

    
if __name__ == "__main__":
    main()


'''
Created on Jul 21, 2020

@author: vladislavkargin
'''
import numpy as np
from scipy import linalg as la
import matplotlib.pyplot as plt
import os

import rgs.pmaps.OrderedTree as ot
import rgs.mndrpy.Pairing as pg
import rgs.mndrpy.Meander as mr

    
def randomDyckPath(n, seed = None):
    '''
     * Generates a random Dyck path with 2n steps
     * 
     * @param n half-size of path
     * @return a sequence of -1 and 1 steps.
    '''
    
    if seed != None:
        np.random.seed(seed)
    path = [1] * n + [-1] * (n + 1)   
    #shuffle randomly
    np.random.shuffle(path)
    
    #do a cyclic shift. 
    #Find the first occurrence of the minimum
    S = np.cumsum(np.array([0] + path))
    '''
    fig1, ax1 = plt.subplots(nrows=1, ncols=1)
    ax1.plot(S)
    ax1.grid(True)
    '''
    m = np.min(S)
    positions = np.where(S == m)[0]
    pos = positions[0]
    #print(pos)
    path1 = path[pos:] + path[:pos]
    del path1[-1]
    return path1

def plotDyckPath(path, ax = None, method = "upperLatticePath"):
    ''' plots a Dyck path. Path is either list of 1, -1 or numpy array
    of 1, - 1'''
    if isinstance(path, list):
        n = int(len(path)/2)
    else:
        n = int(path.shape[0]/2)
    x = 0
    y = 0
    S = [(x, y)]
    if method == "upperLatticePath":
        for i in range(2 * n):
            if path[i] == 1:
                y += 1
            else:
                x += 1
            S.append((x,y))
    else:         
        for i in range(2 * n):
            if path[i] == 1:
                x += 1
            else:
                y += 1
            S.append((x,y))
    #print(S)
    X, Y = list(zip(*S)) 
    #print(X)
    #print(Y)
    if ax == None:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
    major_ticks = np.arange(0, n + 1, 1)
    #minor_ticks = np.arange(0, 101, 5)
    ax.set_xticks(major_ticks)
    #ax.set_xticks(minor_ticks, minor=True)
    ax.set_yticks(major_ticks)
    #ax.set_yticks(minor_ticks, minor=True)
    # And a corresponding grid
    ax.grid(which='both')
    ax.plot(X, Y)
    Z = range(n + 1)
    ax.plot(Z)
    return ax  


def areaDyckPath(path):
    ''' calculate the area of the Dyck path, defined as the number
    of unit squares under the path but above the line y = x'''
    if isinstance(path, list):
        n = int(len(path)/2)
    else:
        n = int(path.shape[0]/2)
    x = 0
    y = 0
    S = [(x, y)]
    for i in range(2 * n):
        if path[i] == 1:
            y += 1
        else:
            x += 1
        S.append((x,y))
    area = 0
    y_prev = 0
    for p in S: #every time as y increases, we add y - x - 1 to the area
        if p[1] > y_prev:
            area += p[1] - p[0]- 1
            y_prev = p[1]
    return area

def areaMeander(mndr):
    '''calculates the area of a meander system defined as the area 
    of the polygone that has the Dyck path for the upper pairing
    as its upper boundary and the reflected Dyck path for the lower
    pairing as its lower boundary. The size of the meander is subtracted
    for normalization, so the area of the standard meander system is 0'''
    area = areaDyckPath(pg.prng2path(mndr.uPairing))
    area += areaDyckPath(pg.prng2path(mndr.dPairing))
    #area += mndr.size()
    return area

def genPairings(n):
    '''This is a generator that yields all (non-crossing) Pairings on [2n].
     Not all of them are different. The algorithm simply generates 2^{2n + 1}
     walks on Z, chooses those that sum to -1 (bridges) 
     and rotate each of them to a Dyck path.
     Then it converts the path to a pairing. 
     Of course it will be difficult to exhaust all paths for large n.
     '''
    y = 0
    width = 2 * n + 1
    while y < 2 ** width: 
        walk = [2 * int(x) - 1 for x in '{:0{size}b}'.format(y,size=width)]
        if sum(walk) != -1:
            y = y + 1
            continue 
        else:
            y = y + 1
            #do a cyclic shift. 
            #Find the first occurrence of the minimum
            S = np.cumsum(np.array([0] + walk))
            m = np.min(S)
            positions = np.where(S == m)[0]
            pos = positions[0]
            walk1 = walk[pos:] + walk[:pos]
            del walk1[-1]
            yield pg.Pairing(path = walk1)
            
def allPairings(n):   
    '''returns a list of all pairings of length 2n'''     
    A = set([])
    for pairing in genPairings(n):
        A.add(pairing)
    return list(A)

def allMeanders(n):
    ''' returns all meander systems of length 2n. '''
    B = []
    A = allPairings(n)
    for p in A:
        for q in A:
            mndr = mr.Meander(p, q)
            B.append(mndr)
    return B

def dot(p, q):
    ''' calculates the dot product of two pairings, which is 
    the number of cycles in the resulting meander.
    Parameters: p and q - pairings'''
    mndr = mr.Meander(p, q)
    cycles, _ = mndr.findCycles()
    c = len(cycles)
    return c

def dotMatrix(n):
    ''' calculate the matrix of all dot products among non-crossing pairings of 
    length 2n.'''
    A = allPairings(n)
    N = len(A)
    M = np.zeros((N, N))
    for i in range(N):
        for j in range(N):
            M[i,j] = dot(A[i], A[j])
    return M

def randomTree(n, w, seed = None):
    '''
    creates a random planar tree with weight sequence w
    '''
    if seed == None:
        seed = 0 #OrderedTree uses different convention 
                # about the seed of Random Generator
    tree = ot.randomTree(n, w, SEED = seed)
    return tree

def randomCrossMeander(n, seed = None):
    '''generate a random crossing meander'''
    if seed != None:
        np.random.seed(seed)
    uP = pg.randomCrossPairing(n)
    dP = pg.randomCrossPairing(n)
    mndr = mr.Meander(uP, dP)
    return mndr


def treeToDyckPath(tree):
    '''converts a planar rooted tree to a Dyck path'''
    A = [-1]* tree.graph.V 
    dyckPath = []
    v = tree.graph.root
    stack = [v]
    A[v] = 0
    #dfsPath = []
    n = 1
    while (len(stack) > 0):
        #print("stack = " + str(stack));
        v = stack[-1]; #peaking at the top of the stack
        #print("processing v = " + str(v))
        flag = False #can we find unprocessed neighbors?
        d = tree.graph.degree(v)
        for i in range(d):
            w = tree.graph.adj[v][i]
            if (A[w] < 0): #this neighbor has not been explored previously
                A[w] = n
                n = n + 1
                stack.append(w)
                #print(dfsPath)
                dyckPath.append(1)
                flag = True
                break #stop searching for an uexplored neighbor
            else: 
                continue
        if (not flag): 
            stack.pop() #no unexplored neighbor around v. Remove it from consideration.
            dyckPath.append(-1)
    del dyckPath[-1]
    return  dyckPath    

def findQuasiClusters(mndr, max_gap_size = 3):
    ''' finds all quasi-cluster cycle in a meander 
    parameter: meander
    returns: list of cluster cycles'''
    cycles, _ = mndr.findCycles()
    clusters = []
    for cycle in cycles:
        support = cycle.copy()
        support.sort()
        isQuasiCluster = True
        for i, x in enumerate(support):
            if i != len(support) - 1:
                gap = support[i + 1] - x
                if gap > max_gap_size:
                    isQuasiCluster = False
                    break
        if isQuasiCluster:
            clusters.append(cycle)
    return clusters

def largestCluster(mndr, max_gap_size = 3):
    ''' finds the largest quasi-cluster cycle in a meander. If there are several, 
    finds and returns one of them'''
    clusters = findQuasiClusters(mndr, max_gap_size = max_gap_size)
    if clusters == []:
        return []
    lengths = [len(cycle) for cycle in clusters]
    i = lengths.index(max(lengths))
    largest = clusters[i]
    return largest        

def largestClusterLength(mndr, max_gap_size = 3):
    largestCycle = largestCluster(mndr, max_gap_size = max_gap_size)
    return len(largestCycle)

def plotRandomProperMeander(n):
    '''gets a random proper meander and plots it'''
    max_iter = 100000 #maximum number of attempts to get a meander 
    counter = 0
    while counter < max_iter:
        mndr = mr.randomMeander(n)
        if mndr.isProper():
            mndr.draw()
            return mndr
        else:
            counter = counter + 1
    print("Max number of attempts: ", max_iter, " exceeded.")
    print("Proper meander is not found")
    return None


def makeMeanderFromCycle(cycle):
    ''' creates a (crossing) meander from a cycle. 
    For example, given a cycle (0, 5, 3, 2, 1, 4), which corresponds to upper arcs
    (0, 5), (3, 2) and (1, 4) and lower arcs (5, 3) (2, 1) and (4, 0)
    creates pairings 
    [5 4 3 2 1 0] and [4 2 1 5 0 3] and the (crossing) meander based on these pairings.
    parameter: cycle, a list of permuted numbers from 0 to 2n - 1 that starts from 0
    returns: a meander
    '''
    if len(cycle)%2 != 0:
        print("Error: the length of the cycle must be even.")
        return 
    n = int(len(cycle)/2)
    uprng = [0] * 2 * n
    dprng = [0] * 2 * n
    for i in range(n):
        uprng[cycle[2 * i]] = cycle[2 * i + 1]
        uprng[cycle[2 * i + 1]] = cycle[2 * i]
        dprng[cycle[2 * i + 1]] = cycle[(2 * i + 2)%(2 * n)]
        dprng[cycle[(2 * i + 2)%(2 * n)]] = cycle[2 * i + 1]
    mndr = mr.Meander(None, None, uprngArray = uprng, dprngArray = dprng)  
    return mndr  
        
'''
For testing methods
'''
def main():
    
    '''
    #Generate and prints a random tree (weighted)
    n = 100
    w = [1, 1, 1]
    tree = randomTree(n, w)
    print(tree)
    tree.draw(drawLabels = True)
    
    
    #Creates a random Dyck path and plot it 
    #path = randomDyckPath(n)
    path = treeToDyckPath(tree)
    #print(path)
    S = np.cumsum([0] + path)
    fig1, ax1 = plt.subplots(nrows=1, ncols=1)
    ax1.plot(S)
    ax1.grid(True)
    ax1.xaxis.set_ticks(np.arange(0,S.shape[0]))
    ax1.yaxis.set_ticks(np.arange(0, np.max(S) + 1))
    
    
    #Generates a random meander, finds a random cluster cycle (all 
    #point in the cycle are near each other and prints them. 
    n = 100    
    mndr = mr.randomMeander(n)  
    fig, ax = mndr.draw()
    clusters = findQuasiClusters(mndr)  
    print("clusters = \n", clusters)
    
    #here it looks for the largest quasi cluster cycle -- some gaps are allowed. 
    mgs = 9
    cycle = largestCluster(mndr, max_gap_size = mgs)
    print("Largest quasi cluster cycle is ", cycle)
    print("Its length is ", largestClusterLength(mndr, max_gap_size= mgs))
    
    mndr.drawCycle(cycle, ax = ax)
    '''    
    
    #let us generate all non-crossing pairings of length n  
    n = 6
    A = allPairings(n)
    print(len(A))

    
    '''
    p = [3, 2, 1, 0, 5, 4]
    q = [1, 0, 3, 2, 5, 4]
    mndr = mr.Meander(p, q)
    mndr.draw()
    x = dot(p, q)
    print(x)
    '''
    
    M = dotMatrix(n)
    print(type(M))
    print(M)
    path = "/Users/vladislavkargin/Dropbox/Programs/forPapers/Meanders/"
    print(os.path.isdir(path))
    np.savetxt(path + "Matrix" + str(n) + ".csv",
               M.astype(int), delimiter = ',')
    w = la.eigvalsh(M)
    print(w)
    
    
    q = 0.5
    #q = 7/8
    Mq = q ** (n - M) 
    print(Mq)
    w = la.eigvalsh(Mq)
    print("Eigenvalues: ", w)
    print("Maximum eigenvalue: ", max(w))
    print("Minimal Eigenvalue: ", min(w))
    print("Minimal Eigenvalue * C_n^2: ", min(w) * (Mq.shape[0]**2))
    print("Number of Negative Eigenvalues: ", (w[w < 0]).shape[0])
    print("Index: ", (w[w > 0]).shape[0] - (w[w < 0]).shape[0])
    plt.plot(w)
    plt.grid()
    
    ''' I used this for an example in my Math 447 class.
    plotRandomProperMeander(5)
    plt.grid(True)        
    '''
    
    '''This is for paper. It is now available as a notebook in Colab'''
    '''
    seed = 3
    n = 5
    path1 = randomDyckPath(n, seed = seed)
    plotDyckPath(path1)
    prng1 = pg.Pairing(path = path1)
    prng1.draw()
    
    area = areaDyckPath(path1)
    print("Area of path1 is ", area)   
     
    path2 = randomDyckPath(n, seed = seed + 2)
    #plotDyckPath(path2, method = "lowerLatticePath")
    area = areaDyckPath(path2)
    print("Area of path2 is ", area)  
    
    prng2 = pg.Pairing(path = path2)
    mndr = mr.Meander(prng1, prng2)
    mndr.draw(drawCycles=True)
    mr.drawAsPolygon(mndr)
   '''
    
    plt.show()
    
    

if __name__ == '__main__':
    main()


'''
Created on May 18, 2020

@author: vladislavkargin

Various methods to calculate statistical properties of various
classes of Graphs, Trees and Maps
'''
#from os.path import dirname, abspath, join
# Find code directory relative to our directory
#THIS_DIR = dirname(__file__)
#CODE_DIR = abspath(join(THIS_DIR, '..', 'code'))

import sys
sys.path.append('..')


import numpy as np
import matplotlib.pyplot as plt
import pmaps.OrderedTree as ot
import pmaps.PlanarGraph as pg
import pmaps.TreeUtility as tu

def probMacroSplit(n, ITER = 100, a = 0.25, SEED = 0, METHOD = "gwTree",
                PARAM = [1, 1, 1]):
    '''
    estimates the probability that a random binary tree on n vertices 
    splits into two pieces of 
    approximately the same size, which means that the size of the smallest 
    piece is > a n
    The probability is conditional on the event that the tree divides at
    the root.
    Arguments
    n - size of the tree
    ITER - number of iterations used to estimate the probability
    a - the threshold used to check if a macro-split occurred. The default
        is 0.25, which means that if the smaller branch is larger than 0.25 n
        the we have a macro-split.  
    SEED - the seed of random generator for replication purposes. If the seed is 
        set to 0 (default) then the results are changing from run to run. 
    METHOD - the method to generate the random tree. If "gwTree" then a random Galton-Watson
        unary-binary tree on n vertices is generated, if "slimTree" then a random slim tree generated
        (a tree with fixed number of vertices and leaves)
    PARAM - the parameter that has to be passed to the generator of the tree
        if METHOD is gwTree, then this is a 3-vector of degree weights (equivalent to 
        offspring probabilities for a critical GW tree). By default
        it is [1, 1, 1]. If METHOD is slimTree then it is the number of leaves in the
        graph.
    '''
    if (SEED != 0):
        np.random.seed(SEED)
        
    #weights = [1, 1, 1]
    counter = 0
    frequency = 0
    while counter < ITER:
        if (METHOD == "qwTree"):
            tree = ot.randomTree(n, PARAM)
        elif (METHOD == "slimTree"):
            tree = ot.randomSlimTree(PARAM, n)
        else:
            print("probMacroSplit says: the METHOD is unknown")
            return
        #print("The degree of root is " + str(len(tree.graph.adj[0])))
        if counter % 10 == 0:
            print("counter = " + str(counter))
        if len(tree.graph.adj[0]) == 2: #the tree is split at the root
            counter = counter + 1
            u = tree.graph.adj[0][0]
            #v = tree.graph.adj[0][1]
            t = tree.sizeVertex(u)/n
            t = np.min([t, 1 - t])
            if t > a:
                frequency = frequency + 1
    return frequency/ITER

def triangulationMaxDegree(n, ITER = 100, SEED = 0):
    '''
    calculate a vector of maximal vertex degrees in random triangulations on n vertices
    The number of replications is given by parameter ITER, which is by default = 100 
    SEED is the seed for the random generator, for replication purposes
    if SEED == 0 (default), then the results change from time to time.
    ''' 
    max_degrees = np.array([0] * ITER)    
    for i in range(ITER):
        if i % 10 == 0:
            print("counter = " + str(i))
        gr = pg.randomTriangulation(n, SEED)
        max_degrees[i] = gr.maxDegree()
    return max_degrees

def plotTriangulationMaxDegrees(start_n = 100, end_n = 1000, step = 100, ITER = 30):
    '''
    creates a plot of average max degrees 
    '''
    max_degrees = [] 
    rng = list(range(start_n, end_n, step))
    print(rng)
    for n in rng:
        print("n = " + str(n))
        mdgr = triangulationMaxDegree(n, ITER)
        max_degrees.append(np.mean(mdgr)/n * np.log(n))
        
    plt.plot(rng, max_degrees, 'ro-', mfc = "blue")
 
    

def main():
    print('Statistics is running')
    
    
    '''
    Here we checking the method probMacroSplit
    '''
    '''
    n = 1000
    a = 0.3
    k = np.int(a * n)
    #prob = probMacroSplit(n, ITER = n)
    prob = probMacroSplit(n, ITER = n, METHOD = "slimTree", PARAM = k)
    print("probability of macro split is " + str(prob))
    exponent = - np.log(prob)/np.log(n)
    print("log(prob)/log(n) = " + str(exponent))
    '''
    
    #plotTriangulationMaxDegrees()
    
    n = 1000
    gr = pg.randomTriangulation(n, method = "decoratedPath")
    

    profile = gr.degreeProfile()
    plt.plot(profile)
    dcut = 21
    plt.xlim(left = 0, right = dcut)
    plt.xlabel("degree (0 corresponds to leafs). cut at d = " + str(dcut))
    plt.ylabel("number of occurrences")
    plt.title("Degree distribution for a triangulation on " + str(n) + " vertices")
    
    
    plt.show()
    
    
    
    print('Statistics has finished')
    

if __name__ == '__main__':
    main()
            
    
    


'''
Created on Jul 25, 2020

@author: vladislavkargin
'''
import rgs.mndrpy.Meander as Meander
import rgs.mndrpy.Utility as ut

import numpy as np
import matplotlib.pyplot as plt

from scipy.stats import skew


def statsLargestCycle(n, ITER = 1000, method = "Uniform", w = None, seed = None):
    '''
    calculate the mean and the variances 
    of the largest cycle in a sample of ITER random meanders
    '''   
    if seed != None:
        np.random.seed(seed) 
    listL = []
    for count in range(ITER):
        if count > 0 and count % 10000 == 0:
            print("statsLargeCycle : iteration " + str(count))
        if method == "Uniform":
            mndr = Meander.randomMeander(n)
        elif method == "Semi":
            mndr = Meander.randomSemiMeander(n)
        elif method == "Weighted":
            mndr = Meander.randomMeander(n, w)
        elif method == "Comb":
            mndr = Meander.randomComb(n)
        elif method == "CrossSemi":
            mndr = Meander.randomCrossSemiMeander(n)
        else: 
            print("StatsLargestCycle says: Unknown type of a random meander")
            return
        lengths = mndr.calcCycleLengths()
        L = max(lengths)
        listL.append(L)
    listL = np.array(listL)
    M = np.mean(listL)
    S = np.std(listL)
    skewness = skew(listL)
    return M, S, skewness, listL

def statsLargestClusterCycle(n, max_gap_size = 3, ITER = 1000):
    '''calculate the mean of the largest cluster cycle'''
    listL = []
    for _ in range(ITER):
        '''
        if count % 1000 == 0:
            print("statsLargestClusterCycle : iteration " + str(count))
        '''
        mndr = Meander.randomMeander(n)
        listL.append(ut.largestClusterLength(mndr,
                            max_gap_size = max_gap_size))
    M = np.mean(listL)
    return M


def statsCycleSpectrum(n, ITER = 10000, method = "Uniform", w = None, seed = None):
    '''calculate the distribution of cycles over cycles lengths
    parameters:
    n = half-length of the meander system, 
    ITER = number of elements in the sample, 
    method = methods for meander generation
    w - pair of weight sequence needed if the method = "Weighted"
    seed - for replication purposes
    returns:
    listL - array of cycle lengths combined for all iterations
    counts - counts of cycle lengths in all iterations 
    '''
    if seed != None:
        np.random.seed(seed) 
    listL = []
    for count in range(ITER):
        if count % 10 == 0:
            print(count)
        if method == "Uniform":
            mndr = Meander.randomMeander(n)
        elif method == "Semi":
            mndr = Meander.randomSemiMeander(n)
        elif method == "Weighted":
            mndr = Meander.randomMeander(n, w)
        elif method == "Comb":
            mndr = Meander.randomComb(n)
        else: 
            print("StatsLargestCycle says: Unknown type of a random meander")
            return
        lengths = mndr.calcCycleLengths()
        listL = listL + lengths
    listL = np.array(listL)
    counts = []
    for i in range(2, np.max(listL) + 2, 2):
        counts.append(np.count_nonzero(listL == i))
    counts = np.array(counts)/ITER
        
    return listL, counts
    
def statsSpacings(n, ITER = 1, method = "Uniform", w = None, seed = None):
    '''
    calculate the statistics of spacings
    of the largest cycle in a sample of ITER random meanders. Each of meanders 
    have length n
    parameters:
    n - length of the meanders. 
    returns:
    S array of spacings in all iterations
    Sp a matched array of spacings from randomly distributed points. 
    M mean of S
    Mp mean of Sp
    '''   
    if seed != None:
        np.random.seed(seed) 
    S = np.array([])
    Sp = np.array([])
    for count in range(ITER):
        if count % 1000 == 0:
            print("statsSpacings : iteration " + str(count))
        if method == "Uniform":
            mndr = Meander.randomMeander(n)
        elif method == "Semi":
            mndr = Meander.randomSemiMeander(n)
        elif method == "Weighted":
            mndr = Meander.randomMeander(n, w)
        elif method == "Comb":
            mndr = Meander.randomComb(n)
        else: 
            print("StatsLargestCycle says: Unknown type of a random meander")
            return
        #mndr.draw()
        (cycles, _) = mndr.findCycles()
        lengths = list(map(lambda x: len(x), cycles))
        L = max(lengths)
        largestC = cycles[lengths.index(L)]
        #print(largestC)
        support = np.sort(largestC)
        L = len(support)
        X = np.arange(2 * n)
        np.random.shuffle(X)
        #print(X)
        X = np.sort(X[:L])
        #print(X)
        spacings = np.diff(support)
        spacingsX = np.diff(X)
        #print(spacingsX)
        S = np.hstack([S, spacings])
        Sp = np.hstack([Sp, spacingsX])
    #calculate means
    M = np.mean(S)
    Mp = np.mean(Sp)
    return S, Sp, M, Mp

def plotMeanAndStdLargestCycle(Nmin, Nmax, ITER, ticks_step = 10, 
                               ymin = 1.6, ymax = 2.3, 
                               methods = ["Uniform"], seed = None):
    
    '''this is a program that creates plots for my meander paper.
    Namely, it creates the plots that show the mean and the std of the largest 
    cycle of a random meander system, normalized by n^\alpha
    parameters:
    Nmin - minimal length of the meander system
    Mmax - maximum length of the meander system
    ymin - lower limit for y-axis
    ymas - upper limit for y-axis
    ITER - number of sample points for meander systems of given length
    method - type of the random meander system
    seed - the seed for random generator if reproducibility is needed. 
    '''
    def normalizer(n):
        if method in ["Uniform", "Semi", "Weighted", "CrossSemi"]:
            return n ** alpha
        elif method == "Comb":
            return np.log2(n) * alpha
        else:
            print("normalizer says: unknown method")
            return 0
    
    '''
    alpha - parameter for normalization
    step - the step between length of meanders
    ticks_step - the step between ticks on the plot
    '''
    
    _, ax1 = plt.subplots(nrows=1, ncols=1)
    ax1.grid(True)
    ax1.xaxis.set_ticks(np.arange(Nmin,Nmax + 1,ticks_step))
    ax1.set_title("Normalized mean of the largest cycle length")
    ax1.set_xlabel("meander system half-length $n$")
    
    _, ax2 = plt.subplots(nrows=1, ncols=1)
    ax2.grid(True)
    ax2.xaxis.set_ticks(np.arange(Nmin, Nmax + 1,ticks_step))
    ax2.set_title("Normalized std of the largest cycle length")
    ax2.set_xlabel("meander system half-length $n$")   
    
    w = ([1,1,1], [1,1,1]) 
    
    for method in methods:
        step = 1
        if method == "Uniform":
            alpha = 4/5
            style = 'r-'
        elif method == "Semi":
            alpha = 4/5
            style = 'b--'
        elif method == "Weighted":
            alpha = 4/5
            step = 10
            style = 'g-.'
        elif method == "Comb":
            alpha = 1 #this is alpha for logarithm
            style = 'r-'
        elif method == "CrossSemi":
            alpha = 1
            style = 'm-'
        
        mLC = []
        sLC = []
        for n in range(Nmin, Nmax + 1, step):
            if n % 10 == 0:
                print(method + " : " + str(n))          
            (M, S, _, _) = statsLargestCycle(n, ITER = ITER, method = method, w = w, seed = seed)
            mLC.append(M/normalizer(n))
            sLC.append(S/normalizer(n))
        mLC = np.array(mLC)
        sLC = np.array(sLC)
        #print(mLC)
        #print(sLC)
    
        ax1.plot(np.arange(Nmin,Nmax + 1, step), mLC, style, label = method)
        ax1.yaxis.set_ticks(np.arange(ymin, ymax, 0.02))
        if method in ["Uniform", "Semi", "Weighted"]:
            ax1.set_ylabel("$\overline{L}/n^{4/5}$")
        elif method in ["CrossSemi"]:
            ax1.set_ylabel("$\overline{L}/n^{1}$")
        else: 
            ax1.set_ylabel("$\overline{L}/\log_2 n$")
        
        ax2.plot(np.arange(Nmin,Nmax + 1,step), sLC, style, label = method)
        if method in ["Uniform", "Semi", "Weighted"]:
            ax2.yaxis.set_ticks(np.arange(ymin - 1, ymax - 1, 0.02))
            ax2.set_ylabel("$\mathrm{std}(L)/n^{4/5}$")
        elif method in ["CrossSemi"]:
            ax2.set_ylabel("$\overline{L}/n^{1}$")
        else: 
            ax2.yaxis.set_ticks(np.arange(ymin - 1.5, ymax - 1.5, 0.02))
            ax2.set_ylabel("$\mathrm{std}(L)/\log_2 n$")
            
    ax1.legend()
    ax2.legend()

def plotLargestClusterSize(Nmin, Nmax, max_gap_size = 3, ITER = 1000,
                            step = 10, tick_step = 100):
    def normalizer(n):
            return np.log2(n)
    mLC = []
    for n in range(Nmin, Nmax + 1, step):
        if n % 100 == 0:
                print("plotLargestClusterSize: iteration " + str(n))          
        M = statsLargestClusterCycle(n, max_gap_size = max_gap_size, ITER = ITER)
        mLC.append(M/normalizer(n))
    mLC = np.array(mLC)
    _, ax1 = plt.subplots(nrows=1, ncols=1)
    ax1.grid(True)
    ax1.xaxis.set_ticks(np.arange(Nmin,Nmax + 1,tick_step))
    ax1.set_title("Normalized mean of the largest cluster cycle length")
    ax1.set_xlabel("meander system half-length $n$")
    ax1.plot(np.arange(Nmin,Nmax + 1, step), mLC)
    
    
def plotCycleSpectrum(n, ITER = 1000, method = "Uniform", seed = None):
    ''' plots the counts of cycle lengths in order to show the dependence of 
    counts on the cycle length
    '''

    lengths, counts = statsCycleSpectrum(n, method = method, ITER = ITER, seed = seed)
    print("Cycle lengths = " + str(lengths) + "\n and counts/n * 8 = " + str(counts * 8 /n))
    #normalize counts: 
    #counts_norm1 = np.log(10**(-7) + counts * 8 /n) #/np.log(np.arange(2, counts.shape[0] + 2))
    #counts_norm2 = np.log(10**(-7) + counts * 8 /n)/np.log(np.arange(2, counts.shape[0] + 2))
    L = counts.shape[0]
    counts_pred = 1.4/(np.arange(1, L + 1)**2)
    
    #plt.plot(counts * 8/n)
    
    _, (ax1, ax2) = plt.subplots(nrows=2, ncols=1)
    ax1.grid(True)
    ax1.plot(counts * 8/n, label = "cycle counts")
    ax1.plot(counts_pred, label = "predicted cycle counts")
    ax1.set_yscale("log")
    #ax2.set_xscale("log", basex = 2)
    ax1.set_title("Counts of cycles with length k normalized by n/8")
    ax1.legend()
    ax1.set_xlabel("$k$")
    ax1.set_ylabel("$8 \overline{Y}/n$")
    
    ax2.grid(True)
    ax2.plot(counts * 8/n, label = "cycle counts")
    ax2.plot(counts_pred, label = "predicted cycle counts")
    ax2.set_yscale("log")
    ax2.set_xscale("log")
    #ax2.set_title("Counts of cycles with length k normalized by n/8")
    ax2.legend()
    ax2.set_xlabel("$k$")
    ax2.set_ylabel("$8 \overline{Y}/n$")
    
    #ax1.plot(counts_norm1) 
    #ax2.plot(counts_norm2)   
    
def histLargestCycleLength(n, ITER = 10000, bins = 30, 
                           method = "Uniform", seed = None):    
    '''plots the histogram of largest cycle length normalized by n^{4/5}'''
    M, S, skewness, lengths = statsLargestCycle(n, ITER = ITER, method = method,
                                                 seed = seed)
    
    print(" M = " + str(M/n**(4/5)) + "; S = " + str(S/n**(4/5)) + "; skewness = " + str(skewness))
    
    plt.hist(lengths/n**(4/5), bins = bins, density = True)
    plt.grid(True)
    plt.title("Density of the normalized largest cycle length")
    plt.xlabel("$L/n^{4/5}$")

def histLargestCycleSpacing(n, beta = 1, ITER = 10000, bins = 30, 
                            method = "Uniform", seed = None):   
    ''' plots the histogram of spacings in a large cycle in comparison 
    with spacings of uniformly distributed points. The spacings are 
    normalized by n^beta where n is the half-length of the meander system.'''
    S, Sp, _, _ = statsSpacings(n, ITER = ITER, method = method, seed = seed)
    #print(S)
    #print(Sp)
    #frequency of short spacings in largest cycle and random points spacings
    N1 = np.count_nonzero(S == 1)
    print("Percentage of spacings of length 1 in largest cycle = " + str(N1/len(S)))
    N3 = np.count_nonzero(S == 3)
    print("Percentage of spacings of length 3 in LC = " + str(N3/len(S)))   
    N5 = np.count_nonzero(S == 5)
    print("Percentage of spacings of length 5 in LC = " + str(N5/len(S)))   
    
    N1p = np.count_nonzero(Sp == 1)
    print("Percentage of spacings of length 1 in matching sample of uniform points = " 
          + str(N1p/len(S)))
    N3p = np.count_nonzero(Sp == 3)
    print("Percentage of spacings of length 3 in MSofUP = " + str(N3p/len(S)))   
    N5p = np.count_nonzero(Sp == 5)
    print("Percentage of spacings of length 5 in MSofUP = " + str(N5p/len(S)))  
    
    Snorm = S/(n ** beta)  
    Spnorm = Sp/(n ** beta)
    _, ax1 = plt.subplots(nrows=1, ncols=1)
    ax1.hist(Spnorm, bins = bins, density = True, label = "Random points spacings")
    ax1.hist(Snorm, bins = bins, density = True, alpha = 0.5, label = "Largest cycle spacings")
    ax1.set_title("Histogram of spacings in the largest cycle")
    ax1.set_xlabel("spacing size $/n$")
    ax1.set_yscale('log', nonposy='clip')
    #ax1.set_xscale('log')
    ax1.legend()
    '''
    ax2.hist(Spnorm, bins = 50, density = True, label = "Random points spacings")
    ax2.hist(Snorm, bins = 50, density = True, alpha = 0.5, label = "Largest cycle spacings")
    #ax2.set_title("Histogram of spacings in the largest cycle")
    ax2.set_xlabel("spacing size $/n^{1/5}$")
    ax2.set_yscale('log', nonposy='clip')
    #ax2.set_xscale('log')
    ax2.legend()
    ''' 
    
'''
For testing methods
'''
def main():
    
    
    '''Plotting a picture of growth in the length of the largest cycle 
    depending on meander half-size n'''
    Nmin = 10
    Nmax = 200  #target 2000
    ticks_step = 100
    ITER = 500  #target 4000
    #methods = ["Uniform", "Semi", "Weighted"]
    methods = ["CrossSemi"]
    
    plotMeanAndStdLargestCycle(Nmin, Nmax, ITER, ticks_step = ticks_step, methods = methods)
    
    '''Plotting the distribution of cycle lengths in a large meander system'''
    #n = 100000
    #ITER = 1000
    #method = "Uniform"
    #plotCycleSpectrum(n, ITER = ITER, method = method)  

    
      
    ''' Plotting the distribution of the largest cycle length in a large meander system'''
    #n = 2000 
    #ITER = 100000 #target 1000000
    #method = "Uniform"
    #bins = 50
    #histLargestCycleLength(n, ITER = ITER, bins = bins, method = method)
 
    
    '''Showing the histogram of the spacings in the largest cycle of a m.s.'''
    #n = 2000 #target 2000
    #ITER = 50000 #target 50000
    #bins = 50
    #histLargestCycleSpacing(n, ITER = ITER, bins = bins)
    
    '''
    #Now let us study how the average spacing grows with n
    Nmin = 100
    Nmax = 6000  #target 2000
    step = 100
    ITER = 400  #target 4000
    method = "Uniform"
    w = None
    seed = None
    ticks_step = 500
    ymin = 0.8
    ymax = 1.2
    style1 = 'r-'
    style2 = 'b--'
    
    _, ax1 = plt.subplots(nrows=1, ncols=1)
    ax1.grid(True)
    ax1.xaxis.set_ticks(np.arange(Nmin,Nmax + 1,ticks_step))
    ax1.set_title("Normalized mean of the largest cycle spacings")
    ax1.set_xlabel("meander system half-length $n$")
    mLC = []
    sLC = []
    for n in range(Nmin, Nmax + 1, step):
        if n % 100 == 0:
            print(method + " : " + str(n))          
        (_, _, M, Mp) = statsSpacings(n, ITER = ITER, method = method,
                                        w = w, seed = seed)
        mLC.append(M/n**(1/5))
        sLC.append(Mp/n**(1/5))
    mLC = np.array(mLC)
    sLC = np.array(sLC)
    #print(mLC)
    #print(sLC)

    ax1.plot(np.arange(Nmin,Nmax + 1, step), mLC, style1, label = "Largest Cycle")
    ax1.plot(np.arange(Nmin,Nmax + 1, step), sLC, style2, label = "Uniform Points")
    ax1.yaxis.set_ticks(np.arange(ymin, ymax, 0.02))
    if method in ["Uniform", "Semi", "Weighted"]:
        ax1.set_ylabel("$\overline{S}/n^{1/5}$")
    else: 
        ax1.set_ylabel("$\overline{S}/\log_2 n$")
    ax1.legend()
    '''
    
    '''
    plots the largest cluster cycle size (normalized)
    '''
    '''
    Nmin = 10
    Nmax = 2000
    ITER = 100
    max_gap_size = 5
    plotLargestClusterSize(Nmin, Nmax, max_gap_size = max_gap_size, 
                           ITER = ITER)
    '''
    
    plt.show()  


if __name__ == '__main__':
    main()